from functools import reduce
import importlib

import django_filters
from django.conf import settings

from apis_core.apis_entities.models import *
from apis_core.apis_vocabularies.models import RelationBaseClass


# The following classes define the filter sets respective to their models.
# Also by what was enabled in the global settings file (or disabled by not explicitley enabling it).
# Hence by default all filters of all model's fields (automatically generated) and all filters manually defined below
# are at first created and then deleted by what was enabled in the settings file
#
# There is a few overrides happening here, which are in order:
# 1.) The filters defined in GenericListFilter
# 2.) The filters automatically defined in model specific ListFilters (by stating exclude = [] in the Meta class)
# 3.) The filters manually defined in model specific ListFilters (by manual field definitions in the Filter class itself)
# If anything is redefined in a step further it overrides the field from the step before
#
# Additionally, in the global settings file:
# The filters defined there can provide a dictionary which can have a "method" or "label" key-value pair
# where then such a key-value from the settings is overriding the respective key-value of a filter defined in this module
# (e.g. using a different method)

# TODO __sresch__ : Turn the logic of returing a filter object into a singleton pattern to avoid redundant instantiations
# TODO __sresch__ : use the order of list of filter fields in settings


#######################################################################
#
#   Generic super class for sharing filters accross all entities
#
#######################################################################

class GenericListFilter(django_filters.FilterSet):

    class Meta:
        model = None # must be overriden with entity specific Filter Classes
        exclude = getattr(settings, "APIS_RELATIONS_FILTER_EXCLUDE", [])

    name = django_filters.CharFilter(method="name_label_filter", label="Name or Label")
    collection = django_filters.ModelMultipleChoiceFilter(queryset=Collection.objects.all())

    # TODO __sresch__ : look into how the date values can be intercepted so that they can be parsed with the same logic as in edit forms
    start_date = django_filters.DateFromToRangeFilter()
    end_date = django_filters.DateFromToRangeFilter()

    # TODO __sresch__ : look into how to change these into auto-complete fields
    related_entity_name = django_filters.CharFilter(method="related_entity_name_filter", label="related entity")
    related_relationtype_name = django_filters.CharFilter(method="related_relationtype_name_filter", label="relationtype")


    def __init__(self, *args, **kwargs):

        # call super init foremost to create dictionary of filters which will be processed further below
        super().__init__(*args, **kwargs)

        def eliminate_unused_filters(default_filter_dict):
            """
            Method to read in from the settings file which filters should be enabled / disabled and if there are
            methods or labels to override the default ones.

            :param default_filter_dict: the default filter dictionary created on filter class instantiation
                (which comprises filters defined: in GenericListFilter, in specific model ListFilter and their defaults)

            :return: a new dictionary which is a subset of the input dictionary and only contains the filters which
                are referenced in the settings file (and if there were methods or labels also referenced, using them)
            """

            enabled_filters = settings.APIS_ENTITIES[self.Meta.model.__name__]["list_filters"]

            filter_dict_tmp = {}

            for enabled_filter in enabled_filters:

                if type(enabled_filter) == str and enabled_filter in default_filter_dict:
                    # If string then just use it, if a filter with such a name is already defined

                    filter_dict_tmp[enabled_filter] = default_filter_dict[enabled_filter]


                elif type(enabled_filter) == dict:
                    # if a dictionary, then look further into if there is a method or label which overrides the defaults

                    enabled_filter_key = list(enabled_filter.keys())[0]

                    if enabled_filter_key in default_filter_dict:

                        # get the dictionary which contains potential method or label overrides
                        enabled_filter_settings_dict = enabled_filter[enabled_filter_key]

                        if "method" in enabled_filter_settings_dict:
                            default_filter_dict[enabled_filter_key].method = enabled_filter_settings_dict["method"]

                        if "label" in enabled_filter_settings_dict:
                            default_filter_dict[enabled_filter_key].label = enabled_filter_settings_dict["label"]

                        filter_dict_tmp[enabled_filter_key] = default_filter_dict[enabled_filter_key]

                else:
                    raise ValueError("Expected either str or dict as type for an individual filter in the settings file.",
                            "\nGot instead:", type(enabled_filter))

            return filter_dict_tmp

        self.filters = eliminate_unused_filters(self.filters)



    def construct_lookup(self, value):
        """
        Parses user input for wildcards and returns a tuple containing the interpreted django lookup string and the trimmed value
        E.g.
            'example' -> ('__icontains', 'example')
            '*example' -> ('__iendswith', 'example')
            'example*' -> ('__istartswith', 'example')
            '"example"' -> ('__iexact', 'example')

        :param value : str : text to be parsed for *
        :return: (lookup : str, value : str)
        """

        if value.startswith("*") and not value.endswith("*"):

            value = value[1:]
            return "__iendswith", value

        elif not value.startswith("*") and value.endswith("*"):

            value = value[:-1]
            return "__istartswith", value

        elif value.startswith('"') and value.endswith('"'):

            value = value[1:-1]
            return "__iexact", value

        else:

            if value.startswith("*") and value.endswith("*"):

                value = value[1:-1]

            return "__icontains", value



    def string_wildcard_filter(self, queryset, name, value):
        lookup, value = self.construct_lookup(value)
        return queryset.filter(**{name + lookup : value})



    def name_label_filter(self, queryset, name, value):
        # TODO __sresch__ : include alternative names queries

        lookup, value = self.construct_lookup(value)

        queryset_related_label=queryset.filter(**{"label__label"+lookup : value})
        queryset_self_name=queryset.filter(**{name+lookup : value})

        return ( queryset_related_label | queryset_self_name ).distinct().all()



    def related_entity_name_filter(self, queryset, name, value):
        """
        Searches through the all name fields of all related entities of a given queryset

        For performance reasons it was not sensible to use django's usual lookup foreign key fields
        (e.g. Person.objects.filter(personwork__related_work__name__icontains=value))
        Because such naive (but convenient) django ORM approach starts from a given model, goes to the related relation,
        goes to the related entity and searches therein. But such lookups create a lot of joins in the underlying sql,
        which get expensive very fast.

        So the following algorithms reverses this order and always selects primary keys which are then used further to look
        for related models.
        Basically what the following does is:
        step 1: looks up the search value in tempentity class (since all entities inherit from there) and selects their primary keys
        step 2: uses the keys from step 1 to filter for related relations and selects their primary keys
        step 3: uses the keys from step 2 to filter for related entities
        step 4: merge the results from step 3 into one queryset and returns this.

        :param queryset: the queryset currently to be filtered on
        :param name: Here not used, because this method filters on names of related entities
        :param value: The value to be filtered for, input by the user or a programmatic call
        :return: filtered queryset
        """

        # step 1
        # first find all tempentities where the lookup and value applies, select only their primary keys.
        lookup, value = self.construct_lookup(value)
        tempentity_hit = TempEntityClass.objects.filter(**{"name" + lookup: value}).values_list("pk", flat=True)

        # list for querysets where relations will be saved which are related to the primary keys of the tempentity list above
        related_relations_to_hit_list = []

        # step 2
        # iterate over every relation related to the current queryset model (e.g for Person: PersonWork, PersonPerson, etc)
        for relation_class in queryset.model.get_related_relation_classes():

            # get the related classes and lookup names of the current relation
            # (e.g. for PersonWork: class Person, class Work, "related_person", "related_work")
            related_entity_classA = relation_class.get_related_entity_classA()
            related_entity_classB = relation_class.get_related_entity_classB()
            related_entity_field_nameA = relation_class.get_related_entity_field_nameA()
            related_entity_field_nameB = relation_class.get_related_entity_field_nameB()

            # Within a relation class, there is two fields which relate to entities, check now which of the two
            # is the same as the current one.
            if queryset.model == related_entity_classA:

                # append filtered relation queryset to list for use later
                related_relations_to_hit_list.append(
                    # filter the current relation for if the other related entity's pk is in the hit list of tempentity class
                    # from before (e.g. if a tempentity's name contains "seu", then its primary key will be used here, to check
                    # if there is a related entity which has the same primary key.
                    relation_class.objects.filter(
                        **{related_entity_field_nameB + "_id__in": tempentity_hit}
                    ).values_list(related_entity_field_nameA + "_id", flat=True)
                )

            # also check the related entity field B (because it can be that both A and B are of the same entity class, e.g. PersonPerson,
            # or that only A is the other one, e.g. when filtering for Work within PersonWork, then the other related class is Person)
            if queryset.model == related_entity_classB:

                # same procedure as above, just with related class and related name being reversed.
                related_relations_to_hit_list.append(
                    relation_class.objects.filter(
                        **{related_entity_field_nameA + "_id__in": tempentity_hit}
                    ).values_list(related_entity_field_nameB + "_id", flat=True)
                )

            # A safety check which should never arise. But if it does, we know there is something wrong with the automated wiring of
            # entities and their respective relation classes (e.g. for Person, there should never be EventWork arising here)
            if queryset.model != related_entity_classA and queryset.model != related_entity_classB:

                raise ValueError("queryset model class has a wrong relation class associated!")

        # step 3
        # Filter the entity queryset for each of the resulting relation querysets produced in the loop before
        queryset_filtered_list = [ queryset.filter(pk__in=related_relation) for related_relation in related_relations_to_hit_list ]

        # step 4
        # merge them all
        result = reduce( lambda a,b : a | b, queryset_filtered_list).distinct()

        return result




    def related_relationtype_name_filter(self, queryset, name, value):
        """
        Searches through the all name fields of all related relationtypes of a given queryset

        The following logic is almost identical to the one in method 'related_entity_name_filter', so please look up its
        comments for documentational purpose therein.

        Differences are commented however.

        :param queryset: the queryset currently to be filtered on
        :param name: Here not used, because this method filters on names of related relationtypes
        :param value: The value to be filtered for, input by the user or a programmatic call
        :return: filtered queryset
        """

        lookup, value = self.construct_lookup(value)

        # look up through name and name_reverse of RelationBaseClass
        relationbaseclass_hit = (
            RelationBaseClass.objects.filter(**{"name" + lookup: value}).values_list("pk", flat=True) |
            RelationBaseClass.objects.filter(**{"name_reverse" + lookup: value}).values_list("pk", flat=True)
        ).distinct()

        related_relations_to_hit_list = []

        for relation_class in queryset.model.get_related_relation_classes():

            related_entity_classA = relation_class.get_related_entity_classA()
            related_entity_classB = relation_class.get_related_entity_classB()
            related_entity_field_nameA = relation_class.get_related_entity_field_nameA()
            related_entity_field_nameB = relation_class.get_related_entity_field_nameB()

            if queryset.model == related_entity_classA:

                # Only difference to method 'related_entity_name_filter' is that the lookup is done on 'relation_type_id'
                related_relations_to_hit_list.append(
                    relation_class.objects.filter(
                        **{"relation_type_id__in": relationbaseclass_hit}
                    ).values_list(related_entity_field_nameA + "_id", flat=True)
                )

            if queryset.model == related_entity_classB:

                # Only difference to method 'related_entity_name_filter' is that the lookup is done on 'relation_type_id'
                related_relations_to_hit_list.append(
                    relation_class.objects.filter(
                        **{"relation_type_id__in": relationbaseclass_hit}
                    ).values_list(related_entity_field_nameB + "_id", flat=True)
                )

            if queryset.model != related_entity_classA and queryset.model != related_entity_classB:

                raise ValueError("queryset model class has a wrong relation class associated!")

        queryset_filtered_list = [ queryset.filter(pk__in=related_relation) for related_relation in related_relations_to_hit_list ]

        result = reduce( lambda a,b : a | b, queryset_filtered_list).distinct()

        return result


    def related_arbitrary_model_name(self, queryset, name, value):
        """
        Searches through an arbitrarily related model on its name field.

        Note that this works only if
            * the related model has a field 'name'
            * the filter using this method has the same name as the field of the model on which the filter is applied.
                (E.g. the field 'profession' on a person relates to another model: the professiontype. Here the filter on a person
                must also be called 'profession' as the field 'profession' exists within the person model and is then used to search in.
                Using this example of professions, such a lookup would be generated: Person.objects.filter(profession__name__... ) )
        """

        lookup, value = self.construct_lookup(value)

        # name variable is the name of the filter and needs the corresponding field within the model
        return queryset.filter( **{ name + "__name" + lookup : value } )



#######################################################################
#
#   Overriding Entity filter classes
#
#######################################################################


class PersonListFilter(GenericListFilter):

    class Meta(GenericListFilter.Meta):
        model = Person

    gender = django_filters.ChoiceFilter(choices=(('', 'any'), ('male', 'male'), ('female', 'female')))
    profession = django_filters.CharFilter(method="related_arbitrary_model_name")
    title = django_filters.CharFilter(method="related_arbitrary_model_name")
    name = django_filters.CharFilter(method="person_name_filter", label="Name or Label of person")

    def person_name_filter(self, queryset, name, value):

        lookup, value = self.construct_lookup(value)

        queryset_related_label=queryset.filter(**{"label__label"+lookup : value})
        queryset_self_name=queryset.filter(**{name+lookup : value})
        queryset_first_name=queryset.filter(**{"first_name"+lookup : value})

        # return QuerySet.union(queryset_related_label, queryset_self_name, queryset_first_name)
        return (queryset_related_label | queryset_self_name | queryset_first_name).distinct().all()

Person.set_list_filter_class(PersonListFilter)


class PlaceListFilter(GenericListFilter):

    class Meta(GenericListFilter.Meta):
        model = Place

    # TODO __sresch__ : decide on margin tolerance of input, for now the number must be precise
    lng = django_filters.NumberFilter(label='Longitude')
    lat = django_filters.NumberFilter(label='Latitude')

Place.set_list_filter_class(PlaceListFilter)


class InstitutionListFilter(GenericListFilter):

    class Meta(GenericListFilter.Meta):
        model = Institution

Institution.set_list_filter_class(InstitutionListFilter)


class EventListFilter(GenericListFilter):

    class Meta(GenericListFilter.Meta):
        model = Event

Event.set_list_filter_class(EventListFilter)


class WorkListFilter(GenericListFilter):

    class Meta(GenericListFilter.Meta):
        model = Work

    kind = django_filters.ModelChoiceFilter(queryset=WorkType.objects.all())

Work.set_list_filter_class(WorkListFilter)


a_ents = getattr(settings, 'APIS_ADDITIONAL_ENTITIES', False)

if a_ents:
    with open(a_ents, 'r') as ents_file:
        ents = yaml.load(ents_file, Loader=yaml.CLoader)
        for ent in ents['entities']:
            ent_class = getattr(importlib.import_module(f"apis_core.apis_entities.models"), ent['name'])
            vocabs = []
            for v in ent.get('vocabs', []):
                ent_class_type = getattr(importlib.import_module(f"apis_core.apis_vocabularies.models"), v)
                vocabs.append((v, ent_class_type))
            class filterMeta:
                model = ent_class
                exclude = GenericListFilter.fields_to_exclude
            dict_fc = {
                "Meta": filterMeta,
            }
            for v in vocabs:
                dict_fc[v] =  django_filters.ModelChoiceFilter(queryset=ent_class_type.objects.all())
            filterclass_ent = type(f"{ent['name'].title()}Filter", (GenericListFilter, ), dict_fc)
            globals()[f"{ent['name'].title()}ListFilter"] = filterclass_ent


def get_list_filter_of_entity(entity):
    """
    Main method to be called somewhere else in the codebase in order to get the FilterClass respective to the entity string input

    :param entity: str: type of entity
    :return: Entity specific FilterClass
    """

    entity_class = AbstractEntity.get_entity_class_of_name(entity)
    filter_class = entity_class.list_filter_class

    if filter_class is not None:
        return filter_class
    else:
        el = entity.title()
        try:
            return globals()[f"{el}ListFilter"]
        except KeyError:
            raise ValueError("Could not find respective filter for given entity type:", el)


from django.db import models

def generate_all_fields():
    """
    This function goes through every entity, relation, and relationtype model and automatically wires them together
    by setting ManyToMany fields to each other through the relation model. This way the relations of a given entity
    to any other entity or relationtype can be queried much more directly and without the overhead of going through
    the relation model each time.

    The wiring is done by going reflectively through the python code and finding in the respective model modules the relevant classes.

    Django's ContentType interface could not be used since this relies on full class declaration before calling any models,
    however since we want to define attributes of models during their declaration, this was not possible and thus python's
    own code inspection had to be used.

    E.g. for Person these fields are auto-generated: event_set, institution_set, personB_set, personA_set, place_set, work_set
    And each of those fields are ManyToMany Managers where their django methods can be used upon, such as all() and filter()

    E.g. for PersonWork these fields are auto-generated: related_person, and related_work.
    for PersonPerson: related_personA, and related_personB

    :return: None
    """

    # TODO __sresch__ : check for best practice on local imports vs circularity problems.
    from apis_core.apis_entities.models import AbstractEntity
    from apis_core.apis_relations.models import AbstractRelation
    from apis_core.apis_vocabularies.models import AbstractRelationType



    def create_function_get_related_entity_class(related_entity_class):
        """
        :param related_entity_class: the class which the generated class method shall return
        :return: an anonymous function which is assigned as method with a label respective to relation and entity classes.
        """
        return classmethod(lambda cls: related_entity_class)

    def create_function_get_related_entity_field_name(related_entity_field_name):
        """
        :param related_entity_field_name: the name of the related entity which the generated class method shall return
        :return: an anonymous function which is assigned as method with a label respective to relation and entity classes.
        """
        return classmethod(lambda cls: related_entity_field_name)




    def pair_up_relations_relationtypes(relation_classes, relationtype_classes):
        """
        helper function to pair up the relation and relationtype classes so that potential mismatches between them are
        detected early on.

        :param relation_classes: classes of all relations from the module apis_relations/models.py
        :param relationtype_classes: classes of all relationtypes from the module apis_vocabularies/models.py
        :return: list of tuples between relations and their respective relationtype classes
        """

        # result list
        rc_rtc_pairs = []

        # helper set, where initially all the classes are inserted, and only removed if a pairing was found
        no_pairing_found = set(relation_classes + relationtype_classes)

        # Iterate over all relation and relationtype classes
        for rc in relation_classes:
            for rtc in relationtype_classes:

                # compare their name, to check if they should be regarded as pair (This depends on the APIS design
                # decision that given a relation class its respective relationtype must be named the same but with a
                # 'Relation' added at the end)
                if rc.__name__ in rtc.__name__:
                    rc_rtc_pairs.append((rc, rtc))
                    no_pairing_found.remove(rc)
                    no_pairing_found.remove(rtc)
                    break

        # Check if any classes are left without any pairing found
        if len(no_pairing_found) > 0:

            message = ""
            for cls in no_pairing_found:
                message += "Found no corresponding Relation or RelationType class to: \n" + str(cls) + "!"

            raise Exception(message)

        return rc_rtc_pairs


    # all relation, relationtype classes to be iterated over:
    relation_relationtype_pairs = pair_up_relations_relationtypes(
        relation_classes=AbstractRelation.get_all_relation_classes(),
        relationtype_classes=AbstractRelationType.get_all_relationtype_classes()
    )
    # all entity classes to be iterated over:
    entity_classes = AbstractEntity.get_all_entity_classes()


    # Iterate over all entity classes, twice, so that all relations between all entity classes are covered.
    for entity_class_a in entity_classes:
        for entity_class_b in entity_classes:

            entity_class_name_a = entity_class_a.__name__.lower()
            entity_class_name_b = entity_class_b.__name__.lower()

            # inner loop iterating over each of the relation_class and relationtype at the same time
            for relation_class, relationtype_class in relation_relationtype_pairs:


                # Check if current relation related to both entities
                # Note that this way two entites are checked twice, such as person-place and place-person
                # but however in the relation model only one of these two exists. Thus the right one will be picked.
                if entity_class_name_a + entity_class_name_b == relation_class.__name__.lower():

                    if entity_class_a != entity_class_b:



                        # On relation models: generate fields from relation to entity and from relation to relationtype
                        ################################################################################################

                        relation_field_name_a = "related_" + entity_class_name_a
                        relation_field_name_b = "related_" + entity_class_name_b

                        # To stick to common Django ORM Syntax, set the field name in the related entity class to
                        # the current relation name and add '_set' to it (since going from the entity one would have
                        # multiple relation instances possibly.)
                        relation_field_name_in_other_class = entity_class_name_a + entity_class_name_b + "_set"

                        # Create the related entity field A.
                        models.ForeignKey(
                            to=entity_class_a,
                            blank=True,
                            null=True,
                            on_delete=models.CASCADE,
                            related_name=relation_field_name_in_other_class
                        ).contribute_to_class(relation_class, relation_field_name_a)

                        # Create the related entity field B.
                        models.ForeignKey(
                            to=entity_class_b,
                            blank=True,
                            null=True,
                            on_delete=models.CASCADE,
                            related_name=relation_field_name_in_other_class
                        ).contribute_to_class(relation_class, relation_field_name_b)

                        # Create the related relationtype field
                        models.ForeignKey(
                            to=relationtype_class,
                            blank=True,
                            null=True,
                            on_delete=models.CASCADE,
                            related_name=relation_field_name_in_other_class
                        ).contribute_to_class(relation_class, "relation_type")

                        # Implemented the following methods programmatically by setting the respective relations,
                        # entity class references, names, etc.
                        # Note that these methods are first defined as stumps within AbstractRelation for documentation
                        # purposes.

                        relation_class.get_related_entity_classA = \
                            create_function_get_related_entity_class(entity_class_a)

                        relation_class.get_related_entity_classB = \
                            create_function_get_related_entity_class(entity_class_b)

                        relation_class.get_related_entity_field_nameA = \
                            create_function_get_related_entity_field_name(relation_field_name_a)

                        relation_class.get_related_entity_field_nameB = \
                            create_function_get_related_entity_field_name(relation_field_name_b)

                        relation_class.add_relation_class_of_entity_class(entity_class_a)
                        relation_class.add_relation_class_of_entity_class(entity_class_b)
                        relation_class.add_relation_field_name_of_entity_class(
                            relation_name=relation_field_name_in_other_class, entity_class=entity_class_a)
                        relation_class.add_relation_field_name_of_entity_class(
                            relation_name=relation_field_name_in_other_class, entity_class=entity_class_b)



                        # On entity models: generate fields from entity to entity and from entity to relationtype
                        ################################################################################################

                        field_name_to_entity_a = entity_class_name_a + "_set"
                        field_name_to_entity_b = entity_class_name_b + "_set"
                        field_name_to_entity_b_relationtype = entity_class_name_b + "_relationtype_set"
                        field_name_to_entity_a_relationtype = entity_class_name_a + "_relationtype_set"

                        # Add those names already into the respective class's list of field names
                        entity_class_a.add_related_entity_field_name(field_name_to_entity_b)
                        entity_class_b.add_related_entity_field_name(field_name_to_entity_a)
                        entity_class_a.add_related_relationtype_field_name(field_name_to_entity_b_relationtype)
                        entity_class_b.add_related_relationtype_field_name(field_name_to_entity_a_relationtype)
                        relationtype_class.add_related_entity_field_name(field_name_to_entity_b)
                        relationtype_class.add_related_entity_field_name(field_name_to_entity_a)

                        # entity A to entity B, and B back to A
                        models.ManyToManyField(
                            to=entity_class_b,
                            through=relation_class,
                            related_name=field_name_to_entity_a,
                            blank=True,
                        ).contribute_to_class(entity_class_a, field_name_to_entity_b)

                        # entity A to RelationType via entity B, and RelationType back to A
                        models.ManyToManyField(
                            to=relationtype_class,
                            through=relation_class,
                            related_name=field_name_to_entity_a,
                            blank=True,
                        ).contribute_to_class(entity_class_a, field_name_to_entity_b_relationtype)

                        # entity B to RelationType via entity A, and RelationType back to B
                        models.ManyToManyField(
                            to=relationtype_class,
                            through=relation_class,
                            related_name=field_name_to_entity_b,
                            blank=True,
                        ).contribute_to_class(entity_class_b, field_name_to_entity_a_relationtype)



                    else:



                        # On relation models: generate fields from relation to entity and from relation to relationtype
                        ################################################################################################

                        # now that both entity classes are the same, just take the values from the class 'a'
                        entity_class_name = entity_class_name_a
                        entity_class = entity_class_a


                        relation_field_name_a = "related_" + entity_class_name + "A"
                        relation_field_name_b = "related_" + entity_class_name + "B"

                        # Within APIS it was manually set that from a given entity which relates to a relation containing
                        # the same entity class twice (e.g. PersonPerson), within the given entity class the
                        # field's name was 'related_<entity>A', and 'related_<entity>B' (e.g. Person.related_personA).
                        # This is not consistent with the other relation fields of a given entity
                        # (e.g. for PersonWork, in Person and Work, it would refer to the relation via 'personwork_set').
                        #
                        # This should be changed, but it would result in crashes where code would not be adapted.
                        #
                        # So, for now, in order to not break APIS and to keep downwards-compatibility, the same fields are
                        # generated as before.
                        relation_field_name_in_other_class_a = relation_field_name_b
                        relation_field_name_in_other_class_b = relation_field_name_a
                        relation_field_name_in_other_class = entity_class_name_a + entity_class_name_b + "_set"

                        # TODO __sresch__ : use the following related name for consistency reasons once most code breaking parts due to this change are identified.
                        #
                        # Unfortuantely it does not seem possible to have an Entity refer to a symmetric relation with
                        # only the relation name (E.g. that both Person A and Person B would refer to the PersonPerson relation
                        # via 'personperson_set'). So later it will be changed to 'personpersonA_set', and 'personpersonB_set'
                        # to be a bit more consistent with other usages. So, uncomment the following once breaking parts
                        # are identified and prepared for it.
                        #
                        # relation_field_name_from_entity_a = entity_class_name * 2 + "B_set"
                        # relation_field_name_from_entity_b = entity_class_name * 2 + "A_set"

                        # Create the related entity field A.
                        models.ForeignKey(
                            to=entity_class,
                            blank=True,
                            null=True,
                            on_delete=models.CASCADE,
                            related_name=relation_field_name_in_other_class_a
                        ).contribute_to_class(relation_class, relation_field_name_a)

                        # Create the related entity field B.
                        models.ForeignKey(
                            to=entity_class,
                            blank=True,
                            null=True,
                            on_delete=models.CASCADE,
                            related_name=relation_field_name_in_other_class_b
                        ).contribute_to_class(relation_class, relation_field_name_b)

                        # Create the related relaiontype field
                        models.ForeignKey(
                            to=relationtype_class,
                            blank=True,
                            null=True,
                            on_delete=models.CASCADE,
                            related_name=relation_field_name_in_other_class
                        ).contribute_to_class(relation_class, "relation_type")

                        # Implemented the following methods programmatically by setting the respective relations,
                        # entity class references, names, etc.
                        # Note that these methods are first defined as stumps within AbstractRelation for documentation
                        # purposes.

                        relation_class.get_related_entity_classA = \
                            create_function_get_related_entity_class(entity_class)

                        relation_class.get_related_entity_classB = \
                            create_function_get_related_entity_class(entity_class)

                        relation_class.get_related_entity_field_nameA = \
                            create_function_get_related_entity_field_name(relation_field_name_a)

                        relation_class.get_related_entity_field_nameB = \
                            create_function_get_related_entity_field_name(relation_field_name_b)

                        relation_class.add_relation_class_of_entity_class(entity_class)
                        relation_class.add_relation_field_name_of_entity_class(
                            relation_name=relation_field_name_in_other_class_a, entity_class=entity_class)
                        relation_class.add_relation_field_name_of_entity_class(
                            relation_name=relation_field_name_in_other_class_b, entity_class=entity_class)



                        # On entity models: generate fields from entity to entity and from entity to relationtype
                        ################################################################################################

                        # Define all the names for the ManyToMany fields generated below
                        field_name_to_entity_a = entity_class_name_a + "A_set"
                        field_name_to_entity_b = entity_class_name_b + "B_set"
                        field_name_to_entity_b_relationtype = entity_class_name_b + "B_relationtype_set"
                        field_name_to_entity_a_relationtype = entity_class_name_a + "A_relationtype_set"

                        # Add those names already into the respective class's list of field names
                        entity_class_a.add_related_entity_field_name(field_name_to_entity_b)
                        entity_class_b.add_related_entity_field_name(field_name_to_entity_a)
                        entity_class_a.add_related_relationtype_field_name(field_name_to_entity_b_relationtype)
                        entity_class_b.add_related_relationtype_field_name(field_name_to_entity_a_relationtype)
                        relationtype_class.add_related_entity_field_name(field_name_to_entity_b)
                        relationtype_class.add_related_entity_field_name(field_name_to_entity_a)

                        # entity A to same entity B, and B back to A
                        models.ManyToManyField(
                            to=entity_class_b,
                            through=relation_class,
                            related_name=field_name_to_entity_a,
                            blank=True,
                            symmetrical=False,
                            through_fields=("related_" + entity_class_name_a + "A", "related_" + entity_class_name_b + "B")
                        ).contribute_to_class(entity_class_a, field_name_to_entity_b)

                        # entity A to RelationType via entity B, and RelationType back to A
                        models.ManyToManyField(
                            to=relationtype_class,
                            through=relation_class,
                            related_name=field_name_to_entity_a,
                            blank=True,
                            symmetrical=False,
                            through_fields=("related_" + entity_class_name_a + "A", "relation_type")
                        ).contribute_to_class(entity_class_a, field_name_to_entity_b_relationtype)

                        # entity B to RelationType via entity A, and RelationType back to B
                        models.ManyToManyField(
                            to=entity_class_b,
                            through=relation_class,
                            related_name=field_name_to_entity_a_relationtype,
                            blank=True,
                            symmetrical=False,
                            through_fields=("relation_type", "related_" + entity_class_name_b + "B")
                        ).contribute_to_class(relationtype_class, field_name_to_entity_b)



                    # if entity_class_a_name + entity_class_b_name == relation_class.__name__.lower()
                    # equals to True, then for entity_class_a and entity_class_b, their respective relation class
                    # has been found, thus interrupt the loop going through these relation classes.
                    break



# generated by appcreator
from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator
from django.urls import reverse, reverse_lazy
from django.views.generic.detail import DetailView
from django.views.generic.edit import DeleteView
from . filters import *
from . forms import *
from . tables import *
from . models import *
from browsing.browsing_utils import GenericListView, BaseCreateView, BaseUpdateView


class TeamMemberListView(GenericListView):

    model = TeamMember
    filter_class = TeamMemberListFilter
    formhelper_class = TeamMemberFilterFormHelper
    table_class = TeamMemberTable
    init_columns = [
        'id', 'description',
    ]


class TeamMemberDetailView(DetailView):

    model = TeamMember
    template_name = 'browsing/generic_detail.html'


class TeamMemberCreate(BaseCreateView):

    model = TeamMember
    form_class = TeamMemberForm

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(TeamMemberCreate, self).dispatch(*args, **kwargs)


class TeamMemberUpdate(BaseUpdateView):

    model = TeamMember
    form_class = TeamMemberForm

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(TeamMemberUpdate, self).dispatch(*args, **kwargs)


class TeamMemberDelete(DeleteView):
    model = TeamMember
    template_name = 'webpage/confirm_delete.html'
    success_url = reverse_lazy('info:teammember_browse')

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(TeamMemberDelete, self).dispatch(*args, **kwargs)


class AboutTheProjectListView(GenericListView):

    model = AboutTheProject
    filter_class = AboutTheProjectListFilter
    formhelper_class = AboutTheProjectFilterFormHelper
    table_class = AboutTheProjectTable
    init_columns = [
        'id', 'description',
    ]


class AboutTheProjectDetailView(DetailView):

    model = AboutTheProject
    template_name = 'browsing/generic_detail.html'


class AboutTheProjectCreate(BaseCreateView):

    model = AboutTheProject
    form_class = AboutTheProjectForm

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(AboutTheProjectCreate, self).dispatch(*args, **kwargs)


class AboutTheProjectUpdate(BaseUpdateView):

    model = AboutTheProject
    form_class = AboutTheProjectForm

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(AboutTheProjectUpdate, self).dispatch(*args, **kwargs)


class AboutTheProjectDelete(DeleteView):
    model = AboutTheProject
    template_name = 'webpage/confirm_delete.html'
    success_url = reverse_lazy('info:about_browse')

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(AboutTheProjectDelete, self).dispatch(*args, **kwargs)


class ProjectInstListView(GenericListView):

    model = ProjectInst
    filter_class = ProjectInstListFilter
    formhelper_class = ProjectInstFilterFormHelper
    table_class = ProjectInstTable
    init_columns = [
        'id', 'description',
    ]


class ProjectInstDetailView(DetailView):

    model = ProjectInst
    template_name = 'browsing/generic_detail.html'


class ProjectInstCreate(BaseCreateView):

    model = ProjectInst
    form_class = ProjectInstForm

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(ProjectInstCreate, self).dispatch(*args, **kwargs)


class ProjectInstUpdate(BaseUpdateView):

    model = ProjectInst
    form_class = ProjectInstForm

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(ProjectInstUpdate, self).dispatch(*args, **kwargs)


class ProjectInstDelete(DeleteView):
    model = ProjectInst
    template_name = 'webpage/confirm_delete.html'
    success_url = reverse_lazy('info:projectinst_browse')

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(ProjectInstDelete, self).dispatch(*args, **kwargs)



import lxml.etree as ET

from django.http import HttpResponse
from django.shortcuts import get_object_or_404

from apis_core.apis_entities.models import Person, Place, Institution, Work
from .tei_utils import get_node_from_template


def person_as_tei(request, pk):
    full = request.GET.get('full')
    res = get_object_or_404(Person, pk=pk)
    doc = get_node_from_template('apis_tei/person.xml', res, full=full)
    tei = ET.tostring(doc, pretty_print=True, encoding='UTF-8')
    return HttpResponse(tei, content_type="application/xml")


def place_as_tei(request, pk):
    full = request.GET.get('full')
    res = get_object_or_404(Place, pk=pk)
    doc = get_node_from_template('apis_tei/place.xml', res, full=full)
    tei = ET.tostring(doc, pretty_print=True, encoding='UTF-8')
    return HttpResponse(tei, content_type="application/xml")


def work_as_tei(request, pk):
    full = request.GET.get('full')
    res = get_object_or_404(Work, pk=pk)
    doc = get_node_from_template('apis_tei/work.xml', res, full=full)
    tei = ET.tostring(doc, pretty_print=True, encoding='UTF-8')
    return HttpResponse(tei, content_type="application/xml")


def org_as_tei(request, pk):
    full = request.GET.get('full')
    res = get_object_or_404(Institution, pk=pk)
    doc = get_node_from_template('apis_tei/org.xml', res, full=full)
    tei = ET.tostring(doc, pretty_print=True, encoding='UTF-8')
    return HttpResponse(tei, content_type="application/xml")





#!/usr/bin/python
# -*- coding: utf-8 -*-
import re

from django.core.validators import RegexValidator

# TODO __sresch__ : remove once sure
date_validator = RegexValidator(regex=re.compile(r'^([0-9]{1,2}\.[0-9]{1,2}\.)?[0-9]{4}$'),
	message="Please enter a valide date: (DD).(MM).YYYY",
	code='invalide_date')
# date_validator2 = RegexValidator(regex=re.compile(r'^([0-9]{1,2}\.[0-9]{1,2}\.)?[0-9]{4}$'),
# 	message="Please enter a valide date: (DD)-(MM)-YYYY.",
# 	code='invalide_date') still needed?


import pandas as pd
from django.db.models import Avg
from rest_framework import serializers

from .utils import calculate_age


class GenericBaseSerializer(serializers.BaseSerializer):

    def to_representation(self, obj):
        data = {
            "items": "some",
            "title": "{}".format('Members by Year'),
            "subtitle": "Person Institution relation type {}".format('some type'),
            "legendy": "legendy",
            "legendx": "legendx",
            "categories": "sorted(dates)",
            "measuredObject": "{} relations".format("find some variable"),
            "ymin": 0,
            "x_axis": "something",
            "payload": "something"
        }
        return data


class AvRelations(GenericBaseSerializer):

    def to_representation(self, obj):
        print(obj[0].__class__.__name__)
        qs = obj.filter(start_date__isnull=False).filter(end_date__isnull=False)
        start_year = int(str(qs.order_by('start_date')[0].start_date)[:4])
        end_year = int(str(qs.order_by('-start_date')[0].start_date)[:4])
        qs = [
            {
                'year': x,
                'members_new': qs.filter(start_date__year=x).count(),
                'members_all': qs.filter(
                    start_date__year__lte=x, end_date__year__gte=x
                ).count(),
            } for x in range(start_year, end_year)
        ]
        df = pd.DataFrame(qs)
        payload = [
            {
                'name': 'all members',
                'data': [x for x in df[['members_all']].values.tolist()]
            },
            {
                'name': 'new members',
                'data': [x for x in df[['members_new']].values.tolist()]
            }
        ]
        data = {
            "items": "some",
            "title": "{}".format('Members by Year'),
            "subtitle": "Person Institution relation type {}".format('some type'),
            "legendy": "legendy",
            "legendx": "legendx",
            "categories": "sorted(dates)",
            "measuredObject": "{} relations".format("find some variable"),
            "ymin": 0,
            "x_axis": df['year'].values.tolist(),
            "payload": payload
        }
        return data


class VisAgeSerializer(serializers.BaseSerializer):

    def to_representation(self, obj):
        qs = [
            {
                'year': x,
                'members_new': obj.filter(start_date__year=x).count(),
                'avg_birth_new': obj.filter(start_date__year=x)
                .aggregate(
                    Avg('related_person__start_date__year')
                )['related_person__start_date__year__avg'],
                'members_all': obj.filter(
                    start_date__year__lte=x, end_date__year__gte=x
                ).count(),
                'avg_birth': obj.filter(start_date__year__lte=x, end_date__year__gte=x)
                .aggregate(
                    Avg('related_person__start_date__year')
                )['related_person__start_date__year__avg']
            } for x in range(1847, 2015)
        ]
        df = pd.DataFrame(qs)
        df['avg_age_new'] = df.apply(lambda row: calculate_age(row, 'avg_birth_new'), axis=1)
        df['avg_age_all'] = df.apply(lambda row: calculate_age(row, 'avg_birth'), axis=1)
        payload = [
            {
                'name': 'average age all',
                'data': [x for x in df[['avg_age_all']].values.tolist()]
            },
            {
                'name': 'average age new',
                'data': [x for x in df[['avg_age_new']].values.tolist()]
            }
        ]
        return payload



# generated by appcreator
import django_filters
from django import forms
from . models import AboutTheProject, TeamMember, ProjectInst


class TeamMemberListFilter(django_filters.FilterSet):

    class Meta:
        model = TeamMember
        fields = "__all__"


class AboutTheProjectListFilter(django_filters.FilterSet):

    class Meta:
        model = AboutTheProject
        fields = "__all__"

class ProjectInstListFilter(django_filters.FilterSet):

    class Meta:
        model = ProjectInst
        fields = "__all__"



from django.conf.urls import url

from .tei_ac import TeiEntAc, TeiCompleterAc

from apis_core.apis_tei import views

app_name = 'apis_tei'

urlpatterns = [
    url(
        r'^autocomplete/(?P<entity>[a-zA-Z0-9-]+)/$',
        TeiEntAc.as_view(),
        name='generic_entities_autocomplete'
    ),
    url(
        r'^tei-completer/(?P<entity>[a-zA-Z0-9-]+)/$',
        TeiCompleterAc.as_view(),
        name='tei_completer_autocomplete'
    ),
    url(
         r'^person/(?P<pk>[0-9]+)$',
        views.person_as_tei,
        name='person_as_tei'
    ),
    url(
        r'^place/(?P<pk>[0-9]+)$',
        views.place_as_tei,
        name='place_as_tei'
    ),
    url(
        r'^org/(?P<pk>[0-9]+)$',
        views.org_as_tei,
        name='org_as_tei'
    ),
    url(
        r'^institution/(?P<pk>[0-9]+)$',
        views.org_as_tei,
        name='org_as_tei'
    ),
    url(
        r'^work/(?P<pk>[0-9]+)$',
        views.work_as_tei,
        name='work_as_tei'
    ),
]



import requests
from django.core.management.base import BaseCommand

from apis_core.apis_vocabularies.models import ProfessionType


class Command(BaseCommand):
    # Show this when the user types help
    help = "fetches profession entities from wiki data and imports them as apis ProfessionType"

    def add_arguments(self, parser):
        parser.add_argument(
            '--lang',
            default="en",
            help='A two digit language code. Defaults to en',
        )

    def handle(self, *args, **options):
        url = 'https://query.wikidata.org/sparql'
        query = """
        SELECT ?item ?itemLabel ?description
        WHERE
        {
          ?item wdt:P31 wd:Q28640.
          ?item schema:description ?description FILTER(lang(?description) = "%s")  .
          SERVICE wikibase:label { bd:serviceParam wikibase:language "%s". }
        }
        """
        lang = options['lang']
        r = requests.get(url, params={'format': 'json', 'query': query % (lang, lang)})
        data = r.json()
        data = [
            [
                x['item']['value'],
                x['itemLabel']['value'],
                x['description']['value']
            ] for x in data['results']['bindings']
        ]
        for x in data:
            if x[1].startswith('Q') or x[1].startswith('L'):
                pass
            else:
                label = f"{x[1]} ({x[0].split('/')[-1]})"
                item, _ = ProfessionType.objects.get_or_create(
                    name=label
                )
                item.description = x[2]
                print(item)



import re

from django.conf import settings
from django.urls import reverse_lazy
from rest_framework import serializers

from .models import Institution, Person, Place, Event, Work
from ..apis_relations.models import PersonInstitution, InstitutionPlace, PersonPlace
from ..apis_vocabularies.models import RelationBaseClass, InstitutionPlaceRelation


class BaseEntitySerializer(serializers.HyperlinkedModelSerializer):
    uri_set = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:uri-detail",
        lookup_field="pk",
        many=True,
        read_only=True,
    )
    collection = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:collection-detail",
        lookup_field="pk",
        many=True,
        read_only=True,
    )
    text = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:text-detail",
        lookup_field="pk",
        many=True,
        read_only=True,
    )


class InstitutionSerializer(BaseEntitySerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:institution-detail", lookup_field="pk"
    )

    kind = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:institutiontype-detail", read_only=True
    )

    class Meta:
        model = Institution
        fields = ("url", "id", "name", "uri_set", "kind", "collection", "text")


class PersonSerializer(BaseEntitySerializer):

    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:person-detail", lookup_field="pk"
    )
    profession = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:professiontype-detail",
        lookup_field="pk",
        many=True,
        read_only=True,
    )

    class Meta:
        model = Person
        fields = (
            "url",
            "id",
            "name",
            "first_name",
            "uri_set",
            "profession",
            "collection",
            "text",
        )


class PlaceSerializer(BaseEntitySerializer):

    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:place-detail", lookup_field="pk"
    )

    kind = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:placetype-detail", lookup_field="pk", read_only=True
    )

    class Meta:
        model = Place
        fields = (
            "url",
            "id",
            "name",
            "uri_set",
            "collection",
            "text",
            "kind",
            "lng",
            "lat",
        )


class EventSerializer(BaseEntitySerializer):

    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:event-detail", lookup_field="pk"
    )

    kind = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:eventtype-detail", lookup_field="pk", read_only=True
    )

    class Meta:
        model = Event
        fields = ("url", "id", "name", "uri_set", "collection", "text", "kind")


class WorkSerializer(BaseEntitySerializer):

    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:work-detail", lookup_field="pk"
    )

    kind = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:worktype-detail", lookup_field="pk", read_only=True
    )

    class Meta:
        model = Work
        fields = ("url", "id", "name", "uri_set", "collection", "text", "kind")


class GeoJsonSerializer(serializers.BaseSerializer):
    def to_representation(self, obj):
        p_pk = self.context.get("p_pk")
        short = False
        url_r = reverse_lazy(
            "apis:apis_entities:resolve_ambigue_place",
            kwargs={"pk": str(p_pk), "uri": obj["id"][7:]},
        )
        long = False
        if "http://www.w3.org/2003/01/geo/wgs84_pos#long" in obj.keys():
            long = float(
                obj["http://www.w3.org/2003/01/geo/wgs84_pos#long"][0]["value"]
            )
            lat = float(obj["http://www.w3.org/2003/01/geo/wgs84_pos#lat"][0]["value"])
        elif "long" in obj.keys():
            long = float(obj["long"][0]["value"])
            lat = float(obj["lat"][0]["value"])
            short = True
        if long:
            popup = ""
            for k in obj.keys():
                if k == "id" or k == "long" or k == "lat":
                    continue
                if not short or k.startswith("http"):
                    title = k.split("#")[-1]
                else:
                    title = k
                popup += "<b>{}:</b> {}<br/>".format(title, obj[k][0]["value"])
            r = {
                "geometry": {"type": "Point", "coordinates": [long, lat]},
                "type": "Feature",
                "properties": {
                    "popupContent": """{}
                    <b>Geonames:</b> <a href='{}'>Select this URI</a>""".format(
                        popup, url_r
                    )
                },
                "id": url_r,
            }
            return r
        else:
            return ""


class GeoJsonSerializerTheme(serializers.BaseSerializer):
    def to_representation(self, obj):
        if obj[0] is None:
            return ""
        url_r = reverse_lazy(
            "apis:apis_core:place-detail", kwargs={"pk": str(obj[0].pk)}
        )
        if obj[0].lng:
            relations = []
            for rel2 in obj[1]:
                if rel2[1] is not None:
                    res_str = f"{rel2[1].name} / {rel2[0].name}"
                else:
                    res_str = f"{rel2[0].name}"
                if rel2[2] is not None:
                    res_str += f" ({rel2[2]}-"
                if rel2[3] is not None:
                    if res_str.endswith("-"):
                        res_str += f"{rel2[3]})"
                    else:
                        res_str += f" (-{rel2[3]})"
                if "(" in res_str and not res_str.endswith(")"):
                    res_str += ")"
                # relations.append((res_str, rel2[2], rel2[3]))
                relations.append(
                    {
                        "id": rel2[-1].pk,
                        "relation": res_str,
                        "start_date": rel2[2],
                        "end_date": rel2[3],
                    }
                )
            r = {
                "geometry": {"type": "Point", "coordinates": [obj[0].lng, obj[0].lat]},
                "type": "Feature",
                "properties": {
                    "name": obj[0].name,
                    "uris": [x.uri for x in obj[0].uri_set.all()],
                    "kind": obj[0].kind.name
                    if obj[0].kind is not None
                    else "undefined",
                    "url": url_r,
                    # "relation_kind": ", ".join([x[0].name for x in obj[1]])
                    "relations": relations,
                },
                "id": url_r,
            }
            return r
        else:
            return ""


class NetJsonEdgeSerializer(serializers.BaseSerializer):
    def to_representation(self, obj):
        ent_obj = obj.__class__.__name__
        ob_match = re.match(r"([A-Z][a-z]+)([A-Z][a-z]+)$", ent_obj)
        rel_a = "related_" + ob_match.group(1).lower()
        rel_b = "related_" + ob_match.group(2).lower()
        if rel_a == rel_b:
            rel_a += "A"
            rel_b += "B"
        r = {
            "source": getattr(obj, rel_a).pk,
            "target": getattr(obj, rel_b).pk,
            "id": obj.pk,
            "type": "arrow",
            "data": dict(),
        }
        r["data"]["start_date"] = obj.start_date_written
        r["data"]["end_date"] = obj.end_date_written
        r["data"]["relation_type"] = obj.relation_type.name

        return r


class NetJsonNodeSerializer(serializers.BaseSerializer):
    def to_representation(self, obj):
        ent_obj = obj.__class__.__name__
        ent_url = reverse_lazy(
            "apis:apis_entities:generic_entities_edit_view",
            kwargs={"pk": str(obj.pk), "entity": ent_obj.lower()},
        )
        tt = """<div class='arrow'></div>
            <div class='sigma-tooltip-header'>{}</div>
            <div class='sigma-tooltip-body'>
            <table>
                <tr><th>Type</th> <td>{}</td></tr>
                <tr><th>Entity</th> <td><a href='{}'>Link</a></td></tr>
            </table>
            <button class='small-button' onclick='expand_node("{}", {})'>expand</button>
            </div>""".format(
            str(obj), ent_obj, ent_url, ent_obj, obj.pk
        )
        r = {
            "type": ent_obj.lower(),
            "label": str(obj),
            "id": obj.pk,
            "tooltip": tt,
            "data": dict(),
        }
        r["data"]["uri"] = [x.uri for x in obj.uri_set.all()]
        r["data"]["collections"] = [x.name for x in obj.collection.all()]
        r["data"]["notes"] = obj.notes
        r["data"]["references"] = obj.references
        r["data"]["start_date"] = obj.start_date_written
        r["data"]["end_date"] = obj.end_date_written
        if ent_obj.lower() != "person":
            if obj.kind:
                r["data"]["kind"] = obj.kind.name
        if ent_obj.lower() == "place":
            r["data"]["lat"] = obj.lat
            r["data"]["lon"] = obj.lng
        if ent_obj.lower() == "person":
            r["data"]["profession"] = [x.name for x in obj.profession.all()]
            if obj.gender:
                r["data"]["gender"] = obj.gender
        return r


class LifePathPlaceSerializer(serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    name = serializers.CharField()
    lat = serializers.FloatField()
    long = serializers.FloatField(source="lng")

    class Meta:
        fields = ["id", "name", "lat", "long"]
        model = Place


class LifePathSerializer(serializers.BaseSerializer):
    place = serializers.SerializerMethodField(method_name="get_place")
    year = serializers.SerializerMethodField(method_name="get_year")
    relation_type = serializers.CharField(source="relation_type__name")

    def get_place(self, obj):
        if isinstance(obj, PersonInstitution):
            inst = obj.related_institution
            rel_type = getattr(settings, "APIS_LOCATED_IN_ATTR", ["situated in",])
            rel_type_inst = getattr(settings, "APIS_INSTITUTION_PRECEEDING", [160, 161])
            ipl_rel = InstitutionPlaceRelation.objects.filter(
                name__in=rel_type
            ).values_list("pk", flat=True)
            plc = InstitutionPlace.objects.filter(
                relation_type_id__in=ipl_rel, related_institution=inst
            )
            if plc.count() == 0:
                inst_lst1 = [
                    x.related_institutionA
                    for x in inst.related_institutionA.filter(
                        relation_type_id__in=rel_type_inst
                    )
                ]
                inst_lst1.extend(
                    [
                        x.related_institutionB
                        for x in inst.related_institutionB.filter(
                            relation_type_id__in=rel_type_inst
                        )
                    ]
                )
                for i2 in inst_lst1:
                    plc = InstitutionPlace.objects.filter(
                        relation_type_id__in=ipl_rel, related_institution=i2
                    )
                    if plc.count() == 1:
                        if plc[0].related_place.lat:
                            break
            if plc.count() == 1:
                plc = plc.first().related_place
                if plc.lng and plc.lat:
                    return LifePathPlaceSerializer(plc).data
        elif isinstance(obj, PersonPlace):
            plc = obj.related_place
            if plc.lat and plc.lng:
                return LifePathPlaceSerializer(plc).data

    def get_year(self, obj):
        if not obj.start_date and not obj.end_date:
            return None
        if obj.start_date and obj.end_date:
            start = int(obj.start_date.strftime("%Y"))
            end = int(obj.end_date.strftime("%Y"))
            return int((start + end) / 2)
        elif obj.start_date:
            return int(obj.start_date.strftime("%Y"))
        elif obj.end_date:
            return int(obj.end_date.strftime("%Y"))

    def to_representation(self, instance):
        p = self.get_place(instance)
        if p is None:
            return None
        res = {
            "id": instance.pk,
            "coords": [p["lat"], p["long"]],
            "name": p["name"],
            "year": self.get_year(instance),
            "relation_type": str(instance.relation_type),
        }
        return res



import json
import os
import pickle

from django.core.management.base import BaseCommand
from django.core.serializers.json import DjangoJSONEncoder

from apis_core.apis_entities.models import AbstractEntity
from apis_core.apis_entities.serializers_generic import EntitySerializer


class Command(BaseCommand):
    help = 'Command to serialize APIS data to json and store the data.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--entity',
            action='store',
            dest='entity',
            default='Person',
            help='Specify the named entity to serialize.',
        )

        parser.add_argument(
            '--output',
            action='store',
            dest='output',
            default=False,
            help='Path of file to store JSON in.',
        )

        parser.add_argument(
            '--only-published',
            action='store_true',
            dest='only-published',
            default=False,
            help='Set if you want to include published relations only. (Boolean, Default: False).',
        )

        parser.add_argument(
            '--filter',
            action='store',
            dest='filter',
            default="{}",
            help='Specify a dictionary of filter arguments for the Entity queryset.',
        )

        parser.add_argument(
            '--use-cache',
            action='store_true',
            dest='use-cache',
            default=False,
            help='Set if you want to use the cached files instead of serializing them egain (Boolean, Default: False).',
        )

        parser.add_argument(
            '--add-texts',
            action='store_true',
            dest='add-texts',
            default=False,
            help='Set if you want to add the texts attached to the entities (Boolean, Default: False).',
        )

    def handle(self, *args, **options):
        ent = AbstractEntity.get_entity_class_of_name(options['entity'])
        res = []
        objcts = ent.objects.filter(**json.loads(options['filter']))
        if objcts.filter(uri__uri__icontains=' ').count() > 0:
            self.stdout.write(self.style.ERROR('URIs found that contain whitespaces'))
            return
        if objcts.count() > 1000 and not options['use-cache']:
            self.stdout.write(self.style.NOTICE('More than 1000 objects, caching'))
            cnt = 0
            while (cnt * 1000) < objcts.count():
                r = []
                for e in objcts[1000*cnt:(1000*cnt+1000)]:
                    r.append(EntitySerializer(e, only_published=options['only-published'], add_texts=options['add-texts']).data)
                with open(f'serializer_cache/{cnt}.pkl', 'wb') as out:
                    pickle.dump(r, out)
                    self.stdout.write(self.style.NOTICE(f'Pickle written to: serializer_cache/{cnt}.pkl'))
                cnt += 1
            res = '/home/sennierer/projects/apis-webpage-base/serializer_cache'
        elif not options['use-cache']:
            for e in objcts:
                res.append(EntitySerializer(e, only_published=options['only-published'], add_texts=options['add-texts']).data)
        elif options['use-cache']:
            self.stdout.write(self.style.NOTICE('using cache for serializing'))
            res = '/home/sennierer/projects/apis-webpage-base/serializer_cache'
        self.stdout.write(self.style.SUCCESS(f'serialized {len(res)} objects'))
        self.stdout.write(self.style.NOTICE('Starting to create the json file'))
        with open(options['output'], 'w+') as outp:
            if isinstance(res, str):
                directory = os.fsencode(res)
                data_lst = []
                for fn in os.listdir(directory):
                    with open(os.path.join(directory, fn), 'rb') as inf:
                        data_lst.extend(pickle.load(inf))
                json.dump(data_lst, outp, cls=DjangoJSONEncoder)
            elif isinstance(res, list):
                json.dump(res, out, pcls=DjangoJSONEncoder)


#!/usr/bin/python
# -*- coding: utf-8 -*-

gn = "http://www.geonames.org/ontology#"
stb_base = 'https://enrich.acdh.oeaw.ac.at/entityhub/site/'
URL_geonames = stb_base + "geoNames_%s/query"
wgs84_pos = "http://www.w3.org/2003/01/geo/wgs84_pos#"
gnd_geo = "http://www.opengis.net/ont/geosparql#"
stb_find = stb_base + u'{}/find'
GenderMappingGND = {'male': 'http://d-nb.info/standards/vocab/gnd/Gender#male',
                    'männlich': 'http://d-nb.info/standards/vocab/gnd/Gender#male',
                    'Mann': 'http://d-nb.info/standards/vocab/gnd/Gender#male',
                    'female': 'http://d-nb.info/standards/vocab/gnd/Gender#female',
                    'weiblich': 'http://d-nb.info/standards/vocab/gnd/Gender#female',
                    'Frau': 'http://d-nb.info/standards/vocab/gnd/Gender#female'}

def date_conversion(date):
    return "{}T00:00:00.000Z".format(date)

autocomp_settings = {
    'score': u'http://stanbol.apache.org/ontology/entityhub/query#score',
    'uri': 'id',
    'label': u'http://www.w3.org/2000/01/rdf-schema#label',
    'Place': [
        {'source': 'Geonames',
        'type': False,
        'url': stb_find.format('geoNames_S_P_A'),
        'fields': {
            'descr': (gn + 'featureCode', 'String'),
            'name': (gn + 'name','String'),
            'long': (wgs84_pos + 'long', 'String'),
            'lat': (wgs84_pos + 'lat', 'String')
        }},
        {'source': 'GeonamesRGN',
        'type': False,
        'url': stb_find.format('geoNames_RGN'),
        'fields': {
            'descr': (gn + 'featureCode','String'),
            'name': (gn + 'name','String'),
            'long': (wgs84_pos + 'long','String'),
            'lat': (wgs84_pos + 'lat','String')
        }},
        {'source': 'GeonamesVAL',
        'type': False,
        'url': stb_find.format('geoNames_VAL'),
        'fields': {
            'descr': (gn + 'featureCode','String'),
            'name': (gn + 'name','String'),
            'long': (wgs84_pos + 'long','String'),
            'lat': (wgs84_pos + 'lat','String')
        }},
        {'source': 'GND',
        'type': u'http://d-nb.info/standards/elementset/gnd#TerritorialCorporateBodyOrAdministrativeUnit',
        'url': stb_find.format('gndTerritorialCorporateBodyOrAdministrativeUnits'),
        'fields': {
            'name': (u'http://d-nb.info/standards/elementset/gnd#preferredNameForThePlaceOrGeographicName','String'),
            'descr': (u'http://d-nb.info/standards/elementset/gnd#definition','String'),
            'long': (gnd_geo + 'asWKT', 'gndLong'),
            'lat': (wgs84_pos + 'lat','String')
        }},
        ],
    'Institution': [{
        'source': 'GND',
        'type': u'http://d-nb.info/standards/elementset/gnd#CorporateBody',
        'url': stb_find.format('gndCorporateBodyAndOrganOfCorporateBody'),
        'fields': {
            'descr': (u'http://d-nb.info/standards/elementset/gnd#definition','String'),
            'name': (u'http://d-nb.info/standards/elementset/gnd#preferredNameForTheCorporateBody','String')}},
            {
        'source': 'GND',
        'type': u'http://d-nb.info/standards/elementset/gnd#OrganOfCorporateBody',
        'url': stb_find.format('gndCorporateBodyAndOrganOfCorporateBody'),
        'fields': {
            'descr': (u'http://d-nb.info/standards/elementset/gnd#definition','String'),
            'name': (u'http://d-nb.info/standards/elementset/gnd#preferredNameForTheCorporateBody','String')}}],
    'Person': [{
        'source': 'GND',
        'type': u'http://d-nb.info/standards/elementset/gnd#DifferentiatedPerson',
        'url': stb_find.format('gndPersons'),
        'search fields': {'gender': ('http://d-nb.info/standards/elementset/gnd#gender',
                                     GenderMappingGND, 'reference'),
                          'start_date': ('http://d-nb.info/standards/elementset/gnd#dateOfBirth',
                                         date_conversion, 'date_exact'),
                          'end_date': ('http://d-nb.info/standards/elementset/gnd#dateOfDeath',
                                         date_conversion, 'date_exact'),
                          'start_date__gt': ('http://d-nb.info/standards/elementset/gnd#dateOfBirth',
                                         date_conversion, 'date_gt'),
                          'start_date__lt': ('http://d-nb.info/standards/elementset/gnd#dateOfBirth',
                                         date_conversion, 'date_lt'),
                          'end_date__gt': ('http://d-nb.info/standards/elementset/gnd#dateOfBirth',
                                         date_conversion, 'date_gt'),
                          'end_date__lt': ('http://d-nb.info/standards/elementset/gnd#dateOfDeath',
                                         date_conversion, 'date_lt')},
        'fields': {
            'descr': (u'http://d-nb.info/standards/elementset/gnd#biographicalOrHistoricalInformation','String'),
            'name': (u'http://d-nb.info/standards/elementset/gnd#preferredNameForThePerson','String'),
            'dateOfBirth': (u'http://d-nb.info/standards/elementset/gnd#dateOfBirth', 'GNDDate'),
            'dateOfDeath': (u'http://d-nb.info/standards/elementset/gnd#dateOfDeath', 'GNDDate')}},
{
        'source': 'GND',
        'type': u'http://enrich.acdh.oeaw.ac.at/stanbol/entityhub/site/gndRoyalOrMemberOfARoyalHouse/',
        'url': stb_find.format('gndRoyalOrMemberOfARoyalHouse'),
        'search fields': {'gender': ('http://d-nb.info/standards/elementset/gnd#gender',
                                     GenderMappingGND, 'reference'),
                          'start_date': ('http://d-nb.info/standards/elementset/gnd#dateOfBirth',
                                         date_conversion, 'date_exact'),
                          'end_date': ('http://d-nb.info/standards/elementset/gnd#dateOfDeath',
                                         date_conversion, 'date_exact'),
                          'start_date__gt': ('http://d-nb.info/standards/elementset/gnd#dateOfBirth',
                                         date_conversion, 'date_gt'),
                          'start_date__lt': ('http://d-nb.info/standards/elementset/gnd#dateOfBirth',
                                         date_conversion, 'date_lt'),
                          'end_date__gt': ('http://d-nb.info/standards/elementset/gnd#dateOfBirth',
                                         date_conversion, 'date_gt'),
                          'end_date__lt': ('http://d-nb.info/standards/elementset/gnd#dateOfDeath',
                                         date_conversion, 'date_lt')},
        'fields': {
            'descr': (u'http://d-nb.info/standards/elementset/gnd#biographicalOrHistoricalInformation','String'),
            'name': (u'http://d-nb.info/standards/elementset/gnd#preferredNameForThePerson','String'),
            'dateOfBirth': (u'http://d-nb.info/standards/elementset/gnd#dateOfBirth', 'GNDDate'),
            'dateOfDeath': (u'http://d-nb.info/standards/elementset/gnd#dateOfDeath', 'GNDDate')}}
    ],
    'Event': [{
        'source': 'GND',
        'type': u'http://d-nb.info/standards/elementset/gnd#HistoricSingleEventOrEra',
        'url': stb_find.format('gndHistoricEvent'),
        'fields': {
            'descr': (u'http://d-nb.info/standards/elementset/gnd#definition','String'),
            'name': (u'http://d-nb.info/standards/elementset/gnd#preferredNameForTheSubjectHeading','String')}}],
    'Work': []
}

geonames_feature_codes = {
    "ADM1": (
        "first-order administrative division",
        "a primary administrative division of a country, such as a state in the United States"),
    "ADM1H": (
        "historical first-order administrative division",
        "a former first-order administrative division"),
    "ADM2": (
        "second-order administrative division",
        "a subdivision of a first-order administrative division"),
    "ADM2H": (
        "historical second-order administrative division",
        "a former second-order administrative division"),
    "ADM3": (
        "third-order administrative division",
        "a subdivision of a second-order administrative division"),
    "ADM3H": (
        "historical third-order administrative division",
        "a former third-order administrative division"),
    "ADM4": (
        "fourth-order administrative division",
        "a subdivision of a third-order administrative division"),
    "ADM4H": (
        "historical fourth-order administrative division",
        "a former fourth-order administrative division"),
    "ADM5": (
        "fifth-order administrative division",
        "a subdivision of a fourth-order administrative division"),
    "ADMD": (
        "administrative division",
        "an administrative division of a country, undifferentiated as to administrative level"),
    "ADMDH": (
        "historical administrative division",
        "a former administrative division of a political entity, \
        undifferentiated as to administrative level"),
    "LTER": (
        "leased area",
        "a tract of land leased to another country, usually for military installations"),
    "PCL": ("political entity", ""),
    "PCLD": ("dependent political entity", ""),
    "PCLF": ("freely associated state", ""),
    "PCLH": ("historical political entity", "a former political entity"),
    "PCLI": ("independent political entity", ""),
    "PCLIX": ("section of independent political entity", ""),
    "PCLS": ("semi-independent political entity", ""),
    "PRSH": ("parish", "an ecclesiastical district"),
    "TERR": ("territory", ""),
    "ZN": ("zone", ""),
    "ZNB": (
        "buffer zone",
        "a zone recognized as a buffer between two nations in which \
        military presence is minimal or absent"),
    "PPL": (
        "populated place",
        "a city, town, village, or other agglomeration of buildings where people live and work"),
    "PPLA": (
        "seat of a first-order administrative division",
        "seat of a first-order administrative division (PPLC takes precedence over PPLA),"),
    "PPLA2": ("seat of a second-order administrative division", ""),
    "PPLA3": ("seat of a third-order administrative division", ""),
    "PPLA4": ("seat of a fourth-order administrative division", ""),
    "PPLC": ("capital of a political entity", ""),
    "PPLCH": (
        "historical capital of a political entity",
        "a former capital of a political entity"),
    "PPLF": (
        "farm village",
        "a populated place where the population is largely engaged in agricultural activities"),
    "PPLG": ("seat of government of a political entity", ""),
    "PPLH": (
        "historical populated place",
        "a populated place that no longer exists"),
    "PPLL": (
        "populated locality",
        "an area similar to a locality but with a small group of dwellings or other buildings"),
    "PPLQ": ("abandoned populated place", ""),
    "PPLR": (
        "religious populated place",
        "a populated place whose population is largely engaged in religious occupations"),
    "PPLS": (
        "populated places",
        "cities, towns, villages, or other agglomerations of buildings where people live and work"),
    "PPLW": (
        "destroyed populated place",
        "a village, town or city destroyed by a natural disaster, or by war"),
    "PPLX": ("section of populated place", ""),
    "STLMT": ("israeli settlement", ""),
    "RGN": ("region", "an area distinguished by one or more observable physical or cultural characteristics")}


class StbGeoQuerySettings:

    def __init__(self, kind='place'):
        self.kind = kind
        self.score = u'http://stanbol.apache.org/ontology/entityhub/query#score'
        self.uri = 'id'
        self.label = u'http://www.w3.org/2000/01/rdf-schema#label'
        self.kind = kind
        self.last_selected = 0

        if kind == 'place':
            self.selected = [gn+'name', gn+'parentPCLI', gn+'parentCountry',
                                gn+'parentADM1', gn+'parentADM2', gn+'parentADM3',
                                gn+'population', gn+'featureCode', wgs84_pos+'lat',
                                wgs84_pos+'long', gn+'alternateName', gn+'officialName',
                                gn+'shortName', gn+'countryCode', gn+'parentFeature']
            self.stored_feature = {
                    'feature': gn+'PPLC',
                    'URL': URL_geonames % 'PPLC'
            }
            self.features = [{
                    'feature': gn+'PPLC',
                    'URL': URL_geonames % 'PPLC'
            },
             {
                    'feature': gn+'PPLA',
                    'URL': URL_geonames % 'PPLA'
            },
                {
                    'feature': gn+'PPLA2',
                    'URL': URL_geonames % 'PPLA2'
            },
                {
                    'feature': gn+'PPLA3',
                    'URL': URL_geonames % 'PPLA3'
            },
                {
                    'feature': gn+'PPLA4',
                    'URL': URL_geonames % 'PPLA4'
            },
                {
                    'feature': gn+'PPL',
                    'URL': URL_geonames % 'PPL'
            }]
        elif kind == 'admin':
            self.selected = [gn+'featureCode']
            self.stored_feature = {
                    'feature': gn+'PCLI',
                    'URL': URL_geonames % 'PCLI'
            }
            self.features = [{
                    'feature': gn+'PCLI',
                    'URL': URL_geonames % 'PCLI'
            },
                {
                    'feature': gn+'ADM1',
                    'URL': URL_geonames % 'ADM1'
            },
                {
                    'feature': gn+'ADM2',
                    'URL': URL_geonames % 'ADM2'
            },
                {
                    'feature': gn+'ADM3',
                    'URL': URL_geonames % 'ADM3'
            }
            ]

    def get_next_feature(self, ft=False):
        if self.last_selected > len(self.features)-1:
            self.stored_feature = False
            return False
        if not ft:
            ft = self.features[self.last_selected]['feature']
        for idnx, x in enumerate(self.features):
            if x['feature'] == ft:
                try:
                    self.last_selected = idnx+1
                    self.stored_feature = self.features[idnx+1]
                    return self.features[idnx+1]
                except:
                    return None
        return self.features[0]

    def get_data(self, query, adm=False):
        if self.kind == 'place' and adm:
            data = {
                'limit': 20, 'constraints': [{
                    'type': 'text',
                    'field': 'http://www.w3.org/2000/01/rdf-schema#label',
                    'text': query},
                    {'type': 'reference', 'field': adm[1], 'value': adm[0]}
                                    ],
                'selected': self.selected
            }
        else:
            data = {
                'limit': 20, 'constraints': [{
                    'type': 'text',
                    'field': 'http://www.w3.org/2000/01/rdf-schema#label',
                    'text': query},
                                    ],
                'selected': self.selected}

        return data



import inspect
import sys

from django.conf import settings


def add_entities(request):
    ent_list = []
    for name, obj in inspect.getmembers(
        sys.modules['apis_core.apis_entities.models'], inspect.isclass
    ):
        if obj.__module__ == 'apis_core.apis_entities.models' and name != "AbstractEntity" and name != "ent_class":
            ent_list.append(str(name).lower())
    res = {
        'entities_list': ent_list,
        'request': request
    }
    return res


def add_relations(request):
    relations_list = []
    for name, obj in inspect.getmembers(
        sys.modules['apis_core.apis_relations.models'], inspect.isclass
    ):
        if obj.__module__ == 'apis_core.apis_relations.models' and name not in ["AbstractRelation", "AnnotationRelationLinkManager", "ent_class", "BaseRelationManager", "RelationPublishedQueryset"]:
            relations_list.append(str(name).lower())
    res = {
        'relations_list': relations_list,
        'request': request
    }
    return res


def add_apis_settings(request):
    """adds the custom settings to the templates"""
    res = {
        'additional_functions': getattr(settings, "APIS_COMPONENTS", []),
        'request': request
    }
    if 'apis_highlighter' in settings.INSTALLED_APPS:
        res['highlighter_active'] = True
    else:
        res['highlighter_active'] = False
    if 'apis_bibsonomy' in settings.INSTALLED_APPS:
        res['bibsonomy_active'] = True
    else:
        res['bibsonomy_active'] = False
    return res



#!/usr/bin/python
# -*- coding: utf-8 -*-
import operator
import re
from functools import reduce

import autocomplete_light.shortcuts as al
import dateutil.parser
import requests
from django.db.models import Q

from apis_core.apis_entities.models import Place, Person, Institution, Event, Work
from apis_core.apis_metainfo.models import Uri
from apis_core.apis_relations.models import PersonPerson, PersonPlace, PersonInstitution
from apis_core.default_settings.NER_settings import autocomp_settings as ac_settings


class StanbolAutocompleteBase(al.AutocompleteListTemplate):
    autocomplete_template = 'apis_templates/autocomplete/stanbol.html'

    widget_attrs = {

    }

    attrs = {
        'data-autocomplete-minimum-characters': 3,
        'placeholder': u'Start typing to get suggestions',
        'class': 'autocomplete-ort-uri form-control'
    }

    @staticmethod
    def parse_stanbol_object(obj, key, *args):
        if len(args) > 0:
            lst1 = args[0]
        else:
            lst1 = None
        if obj[1] == 'GNDDate':
            if lst1 is not None:
                try:
                    return dateutil.parser.parse(lst1[key][0]['value'])
                except:
                    return lst1[key][0]['value']
            else:
                return obj[0]
        elif obj[1] == 'String':
            if lst1 is not None:
                return lst1[key][0]['value']
            else:
                return obj[0]
        elif obj[1] == 'gndLong':
            if lst1 is not None:
                try:
                    return re.search('Point \( [+-]([0-9\.]+) [+-]([0-9\.]+)', lst1[key][0]['value']).group(1)
                except:
                    print('extract fails')
                    return None
            else:
                print('no match')

    def choices_for_request(self):
        ac_type = self.autocomplete_type
        ac_type_model = self.autocomplete_type_model
        choices = []
        headers = {'Content-Type': 'application/json'}
        q = self.request.GET.get('q')
        for m in ac_type_model:
            arg_list = []
            for mm in m[1]:
                arg_list.append(Q(**{mm+'__icontains': q}))
            res = m[0].objects.filter(reduce(operator.or_, arg_list)).distinct()
            for r in res:
                f = dict()
                f['ac_type'] = '_'.join(ac_type)
                f['name'] = r
                f['score'] = None
                try:
                    f['uri'] = Uri.objects.filter(entity=r)[0].uri
                except:
                    continue
                f['source'] = 'db'
                if ac_type[0] == 'Place':
                    if r.lng != None and r.lat != None:
                        f['long'] = str(r.lng)
                        f['lat'] = str(r.lat)
                f['descr'] = m[2][0].format(*[getattr(r, s) for s in m[2][1]])
                choices.append(f)

        for o in ac_type:
            for y in ac_settings[o]:
                ldpath = ""
                for d in y['fields'].keys():
                    ldpath += "{} = <{}>;\n".format(d, y['fields'][d][0])
                data = {
                    'limit': 20,
                    'name': q,
                    'ldpath': ldpath
                }
                try:
                    r = requests.get(y['url'], params=data, headers=headers)
                    if r.status_code != 200:
                        choices.append({'name': 'Connection to Stanbol failed'})
                        continue
                    res = r.json()
                except:
                    choices.append({'name': 'Connection to Stanbol failed'})
                    continue
                for x in res['results']:
                    f = dict()
                    name = x['name'][0]['value']
                    score = str(x[ac_settings['score']][0]['value'])
                    id = x[ac_settings['uri']]
                    f['ac_type'] = '_'.join(ac_type)
                    f['name'] = name
                    f['score'] = score
                    f['uri'] = id
                    f['source'] = y['source']
                    for field in y['fields'].keys():
                        if field in x.keys():
                            f[field] = self.parse_stanbol_object(y['fields'][field], field, x)
                        else:
                            f[field] = None
                    choices.append(f)
        return choices


class AddRelationBaseAutocomplete(al.AutocompleteListTemplate):
    autocomplete_template = 'apis_templates/autocomplete/AddRelation.html'

    widget_attrs = {

    }

    attrs = {
        'data-autocomplete-minimum-characters': 3,
        'placeholder': u'Start typing to get suggestions',
        'class': 'autocomplete-add-relation form-control'
    }

    def choices_for_request(self):
        q = self.request.GET.get('q', None)
        instance_pk = self.request.GET.get('instance_pk', None)
        choices = []
        model_name = self.model2.__name__.lower()
        if instance_pk and q:
            instance = self.model2.objects.get(pk=instance_pk)
        else:
            return choices
        for rel in self.relations:
            if 'related_' + model_name in dir(rel):
                dd = rel.objects.filter(**{
                        'related_' + model_name: instance,
                        'relation_type__name__icontains': q}).exclude(annotation__isnull=False)
                choices.extend(dd)
            elif 'related_' + model_name + 'A' in dir(rel):
                choices.extend(
                    rel.objects.filter(
                        Q(**{
                            'related_' + model_name + 'A': instance,
                            'relation_type__name__icontains': q}) |
                        Q(**{
                            'related_' + model_name + 'B': instance,
                            'relation_type__name__icontains': q})).distinct().exclude(annotation__isnull=False))
        return choices


class PlaceAutocomplete(StanbolAutocompleteBase):
    autocomplete_type = ['Place', ]
    autocomplete_type_model = [(Place, ['name', 'label__label'], ('Status: {}', ['status'])), ]


class InstitutionAutocomplete(StanbolAutocompleteBase):
    autocomplete_type = ['Institution', ]
    autocomplete_type_model = [(Institution, ['name', 'label__label'], ('Status: {}, Gründungsdatum: {}', ['status', 'start_date_written'])), ]


class PersonAutocomplete(StanbolAutocompleteBase):
    autocomplete_type = ['Person', ]
    autocomplete_type_model = [(Person, ['name', 'first_name', 'label__label'], ('Geburtsdatum: {}, Sterbedatum: {}', ['start_date_written', 'end_date_written']))]


class EventAutocomplete(StanbolAutocompleteBase):
    autocomplete_type = ['Event', ]
    autocomplete_type_model = [(Event, ['name', 'label__label'], ('Start date: {}, Status: {}', ['start_date', 'status'])), ]


class WorkAutocomplete(StanbolAutocompleteBase):
    autocomplete_type = ['Work', ]
    autocomplete_type_model = [(Work, ['name', 'label__label'], ('Start date: {}, Status: {}', ['start_date', 'status'])), ]


class AddRelationPersonHighlighterAutocomplete(AddRelationBaseAutocomplete):
    relations = [PersonPerson, PersonPlace, PersonInstitution]
    model2 = Person


al.register(PlaceAutocomplete)
al.register(InstitutionAutocomplete)
al.register(PersonAutocomplete)
al.register(EventAutocomplete)
al.register(WorkAutocomplete)
al.register(AddRelationPersonHighlighterAutocomplete)



import json
import pickle
from rdflib.term import URIRef

import requests
from SPARQLWrapper import SPARQLWrapper, POST, BASIC, JSON
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand
from rdflib import Graph
from rdflib.plugins.memory import IOMemory

from apis_core.apis_entities.api_renderers import EntityToCIDOC
from apis_core.apis_entities.models import AbstractEntity
from apis_core.apis_entities.serializers_generic import EntitySerializer
from apis_core.apis_vocabularies.api_renderers import VocabToSkos
from apis_core.apis_vocabularies.serializers import GenericVocabsSerializer

map_ct = {
    'trig': ('application/x-trig', 'trig'),
    'trix': ('application/trix', 'trix'),
    'xml': ('application/rdf+xml', 'xml'),
    'turtle': ('application/x-turtle', 'ttl'),
    'n3': ('text/rdf+n3', 'n3')
}

base_uri = getattr(settings, 'APIS_BASE_URI', 'http://apis.info')
if base_uri.endswith('/'):
    base_uri = base_uri[:-1]


class Command(BaseCommand):
    help = 'Command to serialize APIS data to cidoc and update a triple store.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--filter',
            action='store',
            dest='filter',
            default="{}",
            help='Specify a dictionary of filter arguments for the Entity queryset.',
        )

        parser.add_argument(
            '--entity',
            action='store',
            dest='entity',
            default='Person',
            help='Specify the named entity to serialize.',
        )

        parser.add_argument(
            '--update',
            action='store_true',
            dest='update',
            default=False,
            help='Use to update triple store (needs credentials).',
        )

        parser.add_argument(
            '--triple-store',
            action='store',
            dest='triplestore',
            default=False,
            help='Specify a tuple of URL, username, password to access triple-store.',
        )

        parser.add_argument(
            '--delete',
            action='store_true',
            dest='delete',
            default=True,
            help='Use to delete the named graph before inserting (defaults to True).',
        )

        parser.add_argument(
            '--named-graph',
            action='store',
            dest='namedgraph',
            default=False,
            help='Uri of named graph to use.',
        )
        
        parser.add_argument(
            '--named-graph-vocabs',
            action='store',
            dest='namedgraph-vocabs',
            default=False,
            help='Uri of named graph to use for serialization of Vovcabs. If this is not set "/vocabs#" is appended to the main named graph',
        )

        parser.add_argument(
            '--output',
            action='store',
            dest='output',
            default=False,
            help='Path of file to store RDF in.',
        )

        parser.add_argument(
            '--update-vocabs',
            action='store_true',
            dest='update-vocabs',
            default=False,
            help='Specify whether to update skosmos vocabularies (needs settings in place).',
        )

        parser.add_argument(
            '--format',
            action='store',
            dest='format',
            default='xml',
            help='Format to use (xml, trig, n3, turtle, nquads, trix).',
        )

        parser.add_argument(
            '--provenance',
            action='store_true',
            dest='provenance',
            default=True,
            help='Specify whether to add void provenance graph (Boolean, Default: True).',
        )

        parser.add_argument(
            '--use-cache',
            action='store_true',
            dest='use-cache',
            default=False,
            help='Set if you want to use the cached files instead of serializing them egain (Boolean, Default: False).',
        )

        parser.add_argument(
            '--include-vocabs',
            action='store_true',
            dest='include-vocabs',
            default=False,
            help='Set if you want to include the vocabs in the entities graph (Boolean, Default: False).',
        )

        parser.add_argument(
            '--only-published',
            action='store_true',
            dest='only-published',
            default=False,
            help='Set if you want to include published relations only. (Boolean, Default: False).',
        )

    def handle(self, *args, **options):
        if options["namedgraph"]:
            named_graph = options["namedgraph"]
        else:
            named_graph = f"{base_uri}/entities#"
        if options["namedgraph-vocabs"]:
            named_graph_vocabs = options["namedgraph-vocabs"]
        else:
            if named_graph[-1] == "#":
                named_graph_vocabs = "/".join(named_graph.split("/")[:-1])+"/vocabs#"
            else:
                named_graph_vocabs = named_graph + "/vocabs#"
        ent = AbstractEntity.get_entity_class_of_name(options['entity'])
        res = []
        objcts = ent.objects.filter(**json.loads(options['filter']))
        if objcts.filter(uri__uri__icontains=' ').count() > 0:
            self.stdout.write(self.style.ERROR('URIs found that contain whitespaces'))
            return
        if objcts.count() > 1000 and not options['use-cache']:
            self.stdout.write(self.style.NOTICE('More than 1000 objects, caching'))
            cnt = 0
            while (cnt * 1000) < objcts.count():
                r = []
                for e in objcts[1000*cnt:(1000*cnt+1000)]:
                    r.append(EntitySerializer(e, only_published=options['only-published']).data)
                with open(f'serializer_cache/{cnt}.pkl', 'wb') as out:
                    pickle.dump(r, out)
                    self.stdout.write(self.style.NOTICE(f'Pickle written to: serializer_cache/{cnt}.pkl'))
                cnt += 1
            res = 'serializer_cache'
        elif not options['use-cache']:
            for e in objcts:
                res.append(EntitySerializer(e, only_published=options['only-published']).data)
        elif options['use-cache']:
            self.stdout.write(self.style.NOTICE('using cache for serializing'))
            res = 'serializer_cache'
        self.stdout.write(self.style.SUCCESS(f'serialized {len(res)} objects'))
        self.stdout.write(self.style.NOTICE('Starting to create the graph'))
        store = IOMemory()
        self.stdout.write(self.style.NOTICE('Starting to create the SKOS vocabs.'))
        graph2 = Graph(
            store, identifier=URIRef(named_graph_vocabs)
        )
        v_res_ser = []
        for v in ContentType.objects.filter(app_label="apis_vocabularies").exclude(model__in=["vocabnames"]):
            v_res = v.model_class().objects.all()
            if v_res.count() > 0:
                v_res_ser.extend(GenericVocabsSerializer(v_res, many=True).data)
        graph2, store = VocabToSkos().render(v_res_ser, g=graph2, store=store)
        fin_vocab = graph2
        fin, store = EntityToCIDOC().render(res, format_1=options['format'], store=store, binary=True, named_graph=named_graph, provenance=options['provenance'])
        if options['output']:
            with open(options['output'], 'wb') as out:
                out.write(fin.serialize(format=options['format']))
            if fin_vocab:
                with open(f"{options['output'].split('.')[0]+'_vocabs.'+options['output'].split('.')[1]}", 'wb') as out2:
                    out2.write(fin_vocab.serialize(format=options['format']))
            self.stdout.write(self.style.SUCCESS(f'Wrote file to {options["output"]}'))
        if options['update-vocabs']:
            vocabs_settings = getattr(settings, 'APIS_SKOSMOS', None)
            if isinstance(vocabs_settings, dict):
                if "triplestore" in vocabs_settings.keys():
                    vocabs_ts_url = vocabs_settings['triplestore']['url']
                    vocabs_ts_user = vocabs_settings['triplestore']['user']
                    vocabs_ts_password = vocabs_settings['triplestore']['password']
                    sparql_serv_skosmos = SPARQLWrapper(f"{vocabs_ts_url}/sparql")
                    sparql_serv_skosmos.setHTTPAuth(BASIC)
                    sparql_serv_skosmos.setCredentials(vocabs_ts_user, vocabs_ts_password)
                    sparql_serv_skosmos.setMethod(POST)
                    sparql_serv_skosmos.setReturnFormat(JSON)
                    if options['delete']:
                        sp_count = f"""
                                SELECT (COUNT(*) AS ?triples)
                                FROM <{named_graph_vocabs}>
                                WHERE {{ ?s ?p ?o }}
                                """
                        sparql_serv_skosmos.setQuery(sp_count)
                        res_count_1 = sparql_serv_skosmos.query().convert()
                        count = int(res_count_1['results']['bindings'][0]['triples']['value'])
                        if count > 0:
                            self.stdout.write(self.style.NOTICE(f'Found {count} triples in skosmos named graph >> deleting'))
                        params = {'graph': f'<{named_graph_vocabs}>'}
                        res3 = requests.delete(f"{vocabs_ts_url}/graph", auth=(vocabs_ts_user, vocabs_ts_password), headers={'Accept': 'application/xml'}, params=params)
                        self.stdout.write(self.style.NOTICE(f'Deleted the graph: {res3.text} {res3.status_code}')) 
                    header = {'Content-Type': map_ct[options['format']][0]}
                    res2 = requests.post(f"{vocabs_ts_url}/graph", headers=header, data=fin_vocab.serialize(format=options['format']), auth=(vocabs_ts_user, vocabs_ts_password))
                    sparql_serv_skosmos.setQuery(sp_count)
                    res_count_1 = sparql_serv_skosmos.query().convert()
                    count = int(res_count_1['results']['bindings'][0]['triples']['value'])
                    if count > 0:
                        self.stdout.write(self.style.NOTICE(f'Found {count} triples after update in store'))
                    if res2.status_code != 200:
                        self.stdout.write(self.style.ERROR(f'Something went wrong when updating: {res2.text}'))
                    else:
                        self.stdout.write(self.style.SUCCESS('Updated the triplestore.'))
                else:
                    self.stdout.write(self.style.ERROR('You need to specify settings for the SKOSMOS triplestore'))
        
        if options['update']:
            print(options['triplestore'])
            if not options['triplestore']:
                url, username, password = getattr(settings, 'APIS_BLAZEGRAPH', (False, False, False))
                if not url:
                    self.stdout.write(self.style.ERROR('When asking for update you need to either specify settings in APIS_BLAZEGRAPH or in the management command (--triple-store)'))
                    return
            else:
                url, username, password = options['triplestore'].split(',')
                if not url or not username or not password:
                    self.stdout.write(self.style.ERROR('Missing some settings for Triplestore update'))
                    return
            sparql_serv = SPARQLWrapper(url)
            sparql_serv.setHTTPAuth(BASIC)
            sparql_serv.setCredentials(username, password)
            sparql_serv.setMethod(POST)
            sparql_serv.setReturnFormat(JSON)
            sp_count = f"""
                    SELECT (COUNT(*) AS ?triples)
                    FROM <{named_graph}>
                    WHERE {{ ?s ?p ?o }}
                    """
            if options['delete']:
                sparql_serv.setQuery(sp_count)
                res_count_1 = sparql_serv.query().convert()
                count = int(res_count_1['results']['bindings'][0]['triples']['value'])
                if count > 0:
                    self.stdout.write(self.style.NOTICE(f'Found {count} triples in named graph >> deleting'))
                params = {'c': f'<{named_graph}>'}
                res3 = requests.delete(url, auth=(username, password), headers={'Accept': 'application/xml'}, params=params)
                self.stdout.write(self.style.NOTICE(f'Deleted the graph: {res3.text} {res3.status_code}'))
                if named_graph_vocabs:
                    params = {'c': f'<{named_graph_vocabs}>'}
                    res4 = requests.delete(url, auth=(username, password), headers={'Accept': 'application/xml'}, params=params)
                    self.stdout.write(self.style.NOTICE(f'Deleted the vocabs graph: {res4.text} {res4.status_code}'))
                for f in ['class', 'property']:
                    sparql_serv.setQuery(f"""
                        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
                        PREFIX void: <http://rdfs.org/ns/void#>


                        DELETE WHERE {{
                            GRAPH <https://omnipot.acdh.oeaw.ac.at/provenance> {{
                                <{named_graph}> void:{f}Partition ?o.
                                ?o ?p ?s
                                }}
                                }}
                    """)
                    res4 = sparql_serv.query().convert()
                sparql_serv.setQuery(f"""
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
                    PREFIX void: <http://rdfs.org/ns/void#>


                    DELETE WHERE {{
                      GRAPH <https://omnipot.acdh.oeaw.ac.at/provenance> {{
                        <{named_graph}> ?p ?o.
                        
                    }}
                    }}
                """)
                res4 = sparql_serv.query()
            header = {'Content-Type': map_ct[options['format']][0]}
            res2 = requests.post(url, headers=header, data=fin.serialize(format=options['format']), auth=(username, password))
            sparql_serv.setQuery(sp_count)
            res_count_1 = sparql_serv.query().convert()
            count = int(res_count_1['results']['bindings'][0]['triples']['value'])
            if count > 0:
                self.stdout.write(self.style.NOTICE(f'Found {count} triples after update in store'))
            if res2.status_code != 200:
                self.stdout.write(self.style.ERROR(f'Something went wrong when updating: {res2.text}'))
            else:
                self.stdout.write(self.style.SUCCESS('Updated the triplestore.'))



from django.test import TestCase
from django.contrib.auth.models import User, Group
from django.urls import reverse

from .models import Person, Event
from apis_core.apis_metainfo.models import Text, Collection
from apis_core.apis_vocabularies.models import ProfessionType
from reversion import revisions as reversion

from datetime import datetime
from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient
from django.contrib.auth.models import Permission


class PersonModelTestCase(TestCase):
    name = "test name"
    first_name = "test first name"
    text = "kjhkjsdhaslkdhadh lasdjalsk alsjdhaksjdh"
    col_name = "Test collection"
    col_name2 = "Test collection2"
    start_date = "1.3.1930"
    end_date = "4.6.1960"
    prof1 = "Lehrer"
    prof2 = "Politiker"

    @classmethod
    def setUpTestData(cls):
        # Set up data for the whole TestCase
        cls.txt = Text.objects.create(text=cls.text)
        cls.col = Collection.objects.create(name=cls.col_name)
        cls.col2 = Collection.objects.create(name=cls.col_name2)
        cls.prof1 = ProfessionType.objects.create(name=cls.prof1)
        cls.prof2 = ProfessionType.objects.create(name=cls.prof2)

    def test_init(self):
        Person()

    def test_create_object(self):
        p = Person.objects.create(
            name=self.name,
            first_name=self.first_name,
            start_date_written=self.start_date,
            end_date_written=self.end_date,
        )

    def test_object_connections(self):
        p = Person.objects.create(
            name=self.name,
            first_name=self.first_name,
            start_date_written=self.start_date,
            end_date_written=self.end_date,
        )
        p.collection.add(self.col)
        p.text.add(self.txt)

    def test_object_validity(self):
        p = Person.objects.create(
            name=self.name,
            first_name=self.first_name,
            start_date_written=self.start_date,
            end_date_written=self.end_date,
        )
        p.collection.add(self.col)
        p.text.add(self.txt)
        self.assertEqual(p.name, self.name)
        self.assertEqual(p.first_name, self.first_name)
        self.assertEqual(p.start_date_written, self.start_date)
        self.assertEqual(p.start_date, datetime.strptime(self.start_date, "%d.%m.%Y"))

    def test_date_parsing_year(self):
        p = Person.objects.create(
            name=self.name,
            first_name=self.first_name,
            start_date_written="1880",
            end_date_written="1890",
        )
        self.assertEqual(p.start_start_date, datetime.strptime("1.1.1880", "%d.%m.%Y"))
        self.assertEqual(p.start_end_date, datetime.strptime("31.12.1880", "%d.%m.%Y"))
        self.assertEqual(p.end_start_date, datetime.strptime("1.1.1890", "%d.%m.%Y"))
        self.assertEqual(p.end_end_date, datetime.strptime("31.12.1890", "%d.%m.%Y"))
        self.assertEqual(p.start_date, datetime.strptime("1.7.1880", "%d.%m.%Y"))

    def test_date_parsing_override(self):
        p = Person.objects.create(
            name=self.name,
            first_name=self.first_name,
            start_date_written="1880<1880-02-01>",
            end_date_written="1890<1890-05-01>",
        )
        self.assertEqual(p.start_date, datetime.strptime("1.2.1880", "%d.%m.%Y"))
        self.assertEqual(p.end_date, datetime.strptime("1.5.1890", "%d.%m.%Y"))

    def test_date_parsing_override_range(self):
        p = Person.objects.create(
            name=self.name,
            first_name=self.first_name,
            start_date_written="1880<1880-05-01,1880-01-01,1880-12-31>",
            end_date_written="1890<1890-05-01>",
        )
        self.assertEqual(p.start_start_date, datetime.strptime("1.1.1880", "%d.%m.%Y"))
        self.assertEqual(p.start_date, datetime.strptime("1.5.1880", "%d.%m.%Y"))
        self.assertEqual(p.start_end_date, datetime.strptime("31.12.1880", "%d.%m.%Y"))

    def test_merge_entities(self):
        p1 = Person.objects.create(name="person1")
        p2 = Person.objects.create(name="person2")
        p1.collection.add(self.col)
        p1.profession.add(self.prof1)
        p2.collection.add(self.col2)
        p2.profession.add(self.prof2)
        p1.merge_with(p2)
        self.assertEqual(p1.collection.all().count(), 2)
        self.assertEqual(p1.profession.all().count(), 2)

    def test_merge_entities_no_selfself(self):
        p1 = Person.objects.first()
        self.assertRaises(ValueError, p1.merge_with, [p1,])
 
"""
    def test_object_reversion(self):
        with reversion.create_revision():
            p = Person.objects.create(
                name=self.name,
                first_name=self.first_name,
                start_date_written=self.start_date,
                end_date_written=self.end_date)
            p.collection.add(self.col)
            p.text.add(self.txt)
        with reversion.create_revision():
            p2 = Person.objects.get(name=self.name)
            p2.name = 'testname 2'
            p2.save()
        self.assertEqual(p2.name, 'testname 2')
        versions = Version.objects.get_for_object(p2)
        self.assertEqual(len(versions), 2)
        self.assertEqual(versions[0].field_dict['name'], 'testname 2')
        self.assertEqual(versions[1].field_dict['name'], self.name)
"""  # TODO: solve the reversion datetime error


class PermissionsModelTestCase(TestCase):
    name = "test name"
    first_name = "test first name"
    text = "kjhkjsdhaslkdhadh lasdjalsk alsjdhaksjdh"
    col_name = "Test collection"
    start_date = "1.3.1930"
    end_date = "4.6.1960"

    @classmethod
    def setUpTestData(cls):
        # Set up data for the whole TestCase
        cls.txt = Text.objects.create(text=cls.text)
        cls.col = Collection.objects.create(name=cls.col_name)
        cls.pers = Person.objects.create(
            name=cls.name,
            first_name=cls.first_name,
            # start_date_written=cls.start_date,  TODO: serializer throwing error on date
            # end_date_written=cls.end_date
        )
        cls.user = User.objects.create_user("testuser", "apisdev16")
        pe = Permission.objects.get(name="Can change person")
        cls.user.user_permissions.add(pe)
        cls.token = Token.objects.create(user=cls.user).key
        cls.group = Group.objects.create(name="testgroup")
        cls.col.groups_allowed.add(cls.group)
        cls.user.groups.add(cls.group)
        cls.c = APIClient()
        cls.c.credentials(HTTP_AUTHORIZATION="Token " + cls.token)

    def test_no_perm(self):
        res = self.c.patch(
            reverse("apis:apis_core:person-detail", kwargs={"pk": self.pers.pk}),
            data={"name": "changed name"},
            # format='json'
        )
        print(f"no permissions, patch: {res.status_code}")
        self.assertEqual(res.status_code, 403)

    def test_no_perm_view(self):
        res = self.c.get(
            reverse("apis:apis_core:person-detail", kwargs={"pk": self.pers.pk}),
        )
        print(f"no permissions, get: {res.status_code}")
        self.assertEqual(res.status_code, 200)

    def test_perm(self):
        self.pers.collection.add(self.col)
        res = self.c.patch(
            reverse("apis:apis_core:person-detail", kwargs={"pk": self.pers.pk}),
            data={"name": "changed name"},
            # format='json'
        )
        print(f"permissions granted, patch: {res.status_code}")
        self.assertEqual(res.status_code, 200)

    def test_perm_removed(self):
        self.pers.collection.remove(self.col)
        res = self.c.patch(
            reverse("apis:apis_core:person-detail", kwargs={"pk": self.pers.pk}),
            data={"name": "changed name"},
            # format='json'
        )
        print(f"permissions revoked, patch: {res.status_code}")
        self.assertEqual(res.status_code, 403)
 


import lxml.etree as ET
from django.conf import settings
from django.template.loader import get_template
from apis_core.apis_entities.models import Work, Place, Person, Institution
from apis_core.apis_relations.models import PlacePlace, PersonWork
from apis_core.apis_vocabularies.models import PersonWorkRelation
from apis_core.helper_functions.utils import get_child_classes 

try:
    birth_rel = settings.BIRTH_REL
except AttributeError:
    birth_rel = False
try:
    death_rel = settings.DEATH_REL
except AttributeError:
    death_rel = False
try:
    pl_a_part_of = settings.PL_A_PART_OF
except AttributeError:
    pl_a_part_of = False
try:
    pl_b_located_in = settings.PL_B_LOCATED_IN
except AttributeError:
    pl_b_located_in = False
try:
    org_located_in = settings.ORG_LOCATED_IN
except AttributeError:
    org_located_in = False
try:
    author_rels = settings.AUTHOR_RELS
except AttributeError:
    author_rels = False

if author_rels:
    author_rels = get_child_classes(author_rels, PersonWorkRelation, labels=False)
else:
    author_rels = False

def get_part_of_relation(res):
    items = []
    for x in PlacePlace.objects.filter(related_placeB=res).filter(relation_type__id__in=pl_b_located_in):
        items.append(x.related_placeA)
    for x in PlacePlace.objects.filter(related_placeA=res).filter(relation_type__id__in=pl_a_part_of):
        items.append(x.related_placeB)
    return list(set(items))

def get_context(res):
    
    context = {}
    context['object'] = res
    context['org_located_in'] = []
    context['birth_rel'] = []
    context['death_rel'] = []
    context['pl_located_in'] = []
    context['author_rels'] = []
    if isinstance(res, Work):
        person_work_ids = get_child_classes(author_rels, PersonWorkRelation, labels=False)
        try:
            context['author_rels'] = PersonWork.objects.filter(relation_type__in=person_work_ids, related_work=res)
        except ValueError:
            context['author_rels'] = []
    if org_located_in and isinstance(res, Institution):
        try:
            context['org_located_in'] = res.institutionplace_set.filter(relation_type__in=org_located_in)
        except AttributeError:
            pass
    if isinstance(res, Person):
        if birth_rel:
            try:
                context['birth_rel'] = res.personplace_set.filter(relation_type__in=birth_rel)
            except AttributeError:
                pass
        if death_rel:
            try:
                context['death_rel'] = res.personplace_set.filter(relation_type__in=death_rel)
            except AttributeError:
                pass
    if isinstance(res, Place):
        if pl_a_part_of and pl_b_located_in:
            try:
                context['pl_located_in'] = get_part_of_relation(res)
            except ValueError:
                pass
    return context


def get_node_from_template(template_path, res, full=True):
    template = get_template(template_path)
    context = get_context(res)
    context['FULL'] = full
    temp_str = f"{template.render(context=context)}"
    node = ET.fromstring(temp_str)
    return node


def tei_header(
    title="ListPerson",
    ent_type="<listPerson/>",
    template_path='apis_tei/tei.xml'
):
    template = get_template(template_path)
    context = {
        "title": title,
        "ent_type": ent_type
    }
    temp_str = f"{template.render(context=context)}"
    node = ET.fromstring(temp_str)
    return node


# Generated by Django 2.2.12 on 2020-06-09 09:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('apis_metainfo', '0004_collection_published'),
    ]

    operations = [
        migrations.AddField(
            model_name='text',
            name='lang',
            field=models.CharField(blank=True, default='deu', help_text="The ISO 639-3 (or 2) code for the label's language.", max_length=3, null=True, verbose_name='ISO Code'),
        ),
    ]



# Generated by Django 2.1.12 on 2020-01-21 12:27

import django.db.models.manager
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('apis_entities', '0001_initial'),
        ('apis_metainfo', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='EventEvent',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='EventWork',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='InstitutionEvent',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='InstitutionInstitution',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='InstitutionPlace',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='InstitutionWork',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='PersonEvent',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='PersonInstitution',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='PersonPerson',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='PersonPlace',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='PersonWork',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='PlaceEvent',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='PlacePlace',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='PlaceWork',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='WorkWork',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
                ('related_workA', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='related_workB', to='apis_entities.Work')),
                ('related_workB', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='related_workA', to='apis_entities.Work')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
            managers=[
                ('annotation_links', django.db.models.manager.Manager()),
            ],
        ),
    ]



from django.contrib.auth.models import User
from django.db.models.query import QuerySet
from rest_framework import serializers

from .models import (
    InstitutionInstitutionRelation, TextType, CollectionType, VocabsBaseClass,
    InstitutionType, ProfessionType, PlaceType, PersonInstitutionRelation, InstitutionPlaceRelation,
    PersonPlaceRelation, VocabNames, PersonPersonRelation, PersonEventRelation, PersonWorkRelation,
    InstitutionEventRelation, InstitutionWorkRelation, PlaceEventRelation, PlaceWorkRelation, PlacePlaceRelation,
    EventWorkRelation, EventEventRelation, WorkWorkRelation, EventType, WorkType, LabelType)


###########################################################
#
#  Generic Vocabs Serializer
#
###########################################################


class GenericVocabsSerializer(serializers.Serializer):
    id = serializers.IntegerField(read_only=True)
    userAdded = serializers.SerializerMethodField(method_name="add_user")
    parent_class = serializers.PrimaryKeyRelatedField(many=False, read_only=True)
    vocab_name = serializers.SerializerMethodField(method_name="add_vocabname")
    name = serializers.CharField()
    label = serializers.CharField()

    def add_user(self, obj):
        print(f"user pk: {obj.userAdded_id}")
        try:
            u = User.objects.get(pk=obj.userAdded_id)
            if u.first_name and u.last_name:
                return f"{u.last_name}, {u.first_name}"
            else:
                return str(u)
        except:
            return ""

    def add_vocabname(self, obj):
        return str(obj.__class__.__name__)

    def __init__(self, *args, **kwargs):
        super(GenericVocabsSerializer, self).__init__(*args, **kwargs)
        if type(self.instance) == QuerySet:
            rt = self.instance[0]
        else:
            rt = self.instance
        if "Relation" in rt.__class__.__name__:
            self.fields['name_reverse'] = serializers.CharField()
            self.fields['label_reverse'] = serializers.CharField()

###########################################################
#
# Meta- Serializers
#
##########################################################


class UserAccSerializer(serializers.ModelSerializer):

    class Meta:
        model = User
        fields = ['username', 'email']


class VocabsBaseSerializer(serializers.HyperlinkedModelSerializer):
    userAdded = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:user-detail",
        lookup_field="pk",
        read_only=True
    )
    parent_class = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:vocabsbaseclass-detail",
        lookup_field="pk",
        read_only=True
    )
    vocab_name = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:vocabnames-detail",
        lookup_field="pk",
        read_only=True
    )


class CollectionTypeSerializer(VocabsBaseSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:collectiontype-detail",
        lookup_field="pk"
    )

    class Meta:
        fields = '__all__'
        model = CollectionType


class VocabsBaseClassSerializer(VocabsBaseSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:vocabsbaseclass-detail",
        lookup_field="pk"
    )

    class Meta:
        fields = '__all__'
        model = VocabsBaseClass


class VocabNamesSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = VocabNames


############################################################
#
# Entity Types
#
###########################################################


class TextTypeSerializer(VocabsBaseSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:texttype-detail",
        lookup_field="pk"
    )
    collections = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:collection-detail",
        lookup_field="pk",
        many=True,
        read_only=True
    )

    class Meta:
        fields = '__all__'
        model = TextType


class InstitutionTypeSerializer(VocabsBaseSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:institutiontype-detail",
        lookup_field="pk"
    )

    class Meta:
        fields = '__all__'
        model = InstitutionType


class ProfessionTypeSerializer(VocabsBaseSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:profession-detail",
        lookup_field="pk"
    )

    class Meta:
        fields = '__all__'
        model = ProfessionType


class PlaceTypeSerializer(VocabsBaseSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:placetype-detail",
        lookup_field="pk"
    )

    class Meta:
        fields = '__all__'
        model = PlaceType


class EventTypeSerializer(VocabsBaseSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:eventtype-detail",
        lookup_field="pk"
    )

    class Meta:
        fields = '__all__'
        model = EventType


class WorkTypeSerializer(VocabsBaseSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:worktype-detail",
        lookup_field="pk"
    )

    class Meta:
        fields = '__all__'
        model = WorkType


class LabelTypeMinimalSerializer(serializers.ModelSerializer):

    class Meta:
        fields = ('id', 'name')
        model = LabelType

#####################################################################
#
# Relations
#
####################################################################


class InstitutionInstitutionRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = InstitutionInstitutionRelation


class PersonPersonRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = PersonPersonRelation


class PersonInstitutionRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = PersonInstitutionRelation


class PersonPlaceRelationSerializer(VocabsBaseSerializer):

    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:personplacerelation-detail",
        lookup_field="pk"
    )

    class Meta:
        fields = '__all__'
        model = PersonPlaceRelation


class PersonEventRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = PersonEventRelation


class PersonWorkRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = PersonWorkRelation


class InstitutionPlaceRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = InstitutionPlaceRelation


class InstitutionEventRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = InstitutionEventRelation


class InstitutionWorkRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = InstitutionWorkRelation


class PlaceEventRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = PlaceEventRelation


class PlaceWorkRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = PlaceWorkRelation


class PlacePlaceRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = PlacePlaceRelation


class EventWorkRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = EventWorkRelation


class EventEventRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = EventEventRelation


class WorkWorkRelationSerializer(VocabsBaseSerializer):

    class Meta:
        fields = '__all__'
        model = WorkWorkRelation



# Generated by Django 2.1.12 on 2020-01-21 12:27

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('apis_vocabularies', '0001_initial'),
        ('apis_labels', '0001_initial'),
        ('apis_metainfo', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='label',
            name='label_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='apis_vocabularies.LabelType'),
        ),
        migrations.AddField(
            model_name='label',
            name='temp_entity',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='apis_metainfo.TempEntityClass'),
        ),
    ]



# Generated by Django 2.1.12 on 2020-01-21 12:27

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('apis_vocabularies', '0001_initial'),
        ('apis_entities', '0001_initial'),
        ('apis_relations', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='work',
            name='event_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='work_set', through='apis_relations.EventWork', to='apis_vocabularies.EventWorkRelation'),
        ),
        migrations.AddField(
            model_name='work',
            name='institution_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='work_set', through='apis_relations.InstitutionWork', to='apis_vocabularies.InstitutionWorkRelation'),
        ),
        migrations.AddField(
            model_name='work',
            name='kind',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='apis_vocabularies.WorkType'),
        ),
        migrations.AddField(
            model_name='work',
            name='person_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='work_set', through='apis_relations.PersonWork', to='apis_vocabularies.PersonWorkRelation'),
        ),
        migrations.AddField(
            model_name='work',
            name='place_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='work_set', through='apis_relations.PlaceWork', to='apis_vocabularies.PlaceWorkRelation'),
        ),
        migrations.AddField(
            model_name='work',
            name='workB_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='workA_set', through='apis_relations.WorkWork', to='apis_vocabularies.WorkWorkRelation'),
        ),
        migrations.AddField(
            model_name='work',
            name='workB_set',
            field=models.ManyToManyField(blank=True, related_name='workA_set', through='apis_relations.WorkWork', to='apis_entities.Work'),
        ),
        migrations.AddField(
            model_name='place',
            name='event_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='place_set', through='apis_relations.PlaceEvent', to='apis_vocabularies.PlaceEventRelation'),
        ),
        migrations.AddField(
            model_name='place',
            name='event_set',
            field=models.ManyToManyField(blank=True, related_name='place_set', through='apis_relations.PlaceEvent', to='apis_entities.Event'),
        ),
        migrations.AddField(
            model_name='place',
            name='institution_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='place_set', through='apis_relations.InstitutionPlace', to='apis_vocabularies.InstitutionPlaceRelation'),
        ),
        migrations.AddField(
            model_name='place',
            name='kind',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='apis_vocabularies.PlaceType'),
        ),
        migrations.AddField(
            model_name='place',
            name='person_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='place_set', through='apis_relations.PersonPlace', to='apis_vocabularies.PersonPlaceRelation'),
        ),
        migrations.AddField(
            model_name='place',
            name='placeB_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='placeA_set', through='apis_relations.PlacePlace', to='apis_vocabularies.PlacePlaceRelation'),
        ),
        migrations.AddField(
            model_name='place',
            name='placeB_set',
            field=models.ManyToManyField(blank=True, related_name='placeA_set', through='apis_relations.PlacePlace', to='apis_entities.Place'),
        ),
        migrations.AddField(
            model_name='place',
            name='work_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='place_set', through='apis_relations.PlaceWork', to='apis_vocabularies.PlaceWorkRelation'),
        ),
        migrations.AddField(
            model_name='place',
            name='work_set',
            field=models.ManyToManyField(blank=True, related_name='place_set', through='apis_relations.PlaceWork', to='apis_entities.Work'),
        ),
        migrations.AddField(
            model_name='person',
            name='event_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='person_set', through='apis_relations.PersonEvent', to='apis_vocabularies.PersonEventRelation'),
        ),
        migrations.AddField(
            model_name='person',
            name='event_set',
            field=models.ManyToManyField(blank=True, related_name='person_set', through='apis_relations.PersonEvent', to='apis_entities.Event'),
        ),
        migrations.AddField(
            model_name='person',
            name='institution_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='person_set', through='apis_relations.PersonInstitution', to='apis_vocabularies.PersonInstitutionRelation'),
        ),
        migrations.AddField(
            model_name='person',
            name='institution_set',
            field=models.ManyToManyField(blank=True, related_name='person_set', through='apis_relations.PersonInstitution', to='apis_entities.Institution'),
        ),
        migrations.AddField(
            model_name='person',
            name='personB_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='personA_set', through='apis_relations.PersonPerson', to='apis_vocabularies.PersonPersonRelation'),
        ),
        migrations.AddField(
            model_name='person',
            name='personB_set',
            field=models.ManyToManyField(blank=True, related_name='personA_set', through='apis_relations.PersonPerson', to='apis_entities.Person'),
        ),
        migrations.AddField(
            model_name='person',
            name='place_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='person_set', through='apis_relations.PersonPlace', to='apis_vocabularies.PersonPlaceRelation'),
        ),
        migrations.AddField(
            model_name='person',
            name='place_set',
            field=models.ManyToManyField(blank=True, related_name='person_set', through='apis_relations.PersonPlace', to='apis_entities.Place'),
        ),
        migrations.AddField(
            model_name='person',
            name='profession',
            field=models.ManyToManyField(blank=True, to='apis_vocabularies.ProfessionType'),
        ),
        migrations.AddField(
            model_name='person',
            name='title',
            field=models.ManyToManyField(blank=True, to='apis_vocabularies.Title'),
        ),
        migrations.AddField(
            model_name='person',
            name='work_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='person_set', through='apis_relations.PersonWork', to='apis_vocabularies.PersonWorkRelation'),
        ),
        migrations.AddField(
            model_name='person',
            name='work_set',
            field=models.ManyToManyField(blank=True, related_name='person_set', through='apis_relations.PersonWork', to='apis_entities.Work'),
        ),
        migrations.AddField(
            model_name='institution',
            name='event_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='institution_set', through='apis_relations.InstitutionEvent', to='apis_vocabularies.InstitutionEventRelation'),
        ),
        migrations.AddField(
            model_name='institution',
            name='event_set',
            field=models.ManyToManyField(blank=True, related_name='institution_set', through='apis_relations.InstitutionEvent', to='apis_entities.Event'),
        ),
        migrations.AddField(
            model_name='institution',
            name='institutionB_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='institutionA_set', through='apis_relations.InstitutionInstitution', to='apis_vocabularies.InstitutionInstitutionRelation'),
        ),
        migrations.AddField(
            model_name='institution',
            name='institutionB_set',
            field=models.ManyToManyField(blank=True, related_name='institutionA_set', through='apis_relations.InstitutionInstitution', to='apis_entities.Institution'),
        ),
        migrations.AddField(
            model_name='institution',
            name='kind',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='apis_vocabularies.InstitutionType'),
        ),
        migrations.AddField(
            model_name='institution',
            name='person_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='institution_set', through='apis_relations.PersonInstitution', to='apis_vocabularies.PersonInstitutionRelation'),
        ),
        migrations.AddField(
            model_name='institution',
            name='place_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='institution_set', through='apis_relations.InstitutionPlace', to='apis_vocabularies.InstitutionPlaceRelation'),
        ),
        migrations.AddField(
            model_name='institution',
            name='place_set',
            field=models.ManyToManyField(blank=True, related_name='institution_set', through='apis_relations.InstitutionPlace', to='apis_entities.Place'),
        ),
        migrations.AddField(
            model_name='institution',
            name='work_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='institution_set', through='apis_relations.InstitutionWork', to='apis_vocabularies.InstitutionWorkRelation'),
        ),
        migrations.AddField(
            model_name='institution',
            name='work_set',
            field=models.ManyToManyField(blank=True, related_name='institution_set', through='apis_relations.InstitutionWork', to='apis_entities.Work'),
        ),
        migrations.AddField(
            model_name='event',
            name='eventB_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='eventA_set', through='apis_relations.EventEvent', to='apis_vocabularies.EventEventRelation'),
        ),
        migrations.AddField(
            model_name='event',
            name='eventB_set',
            field=models.ManyToManyField(blank=True, related_name='eventA_set', through='apis_relations.EventEvent', to='apis_entities.Event'),
        ),
        migrations.AddField(
            model_name='event',
            name='institution_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='event_set', through='apis_relations.InstitutionEvent', to='apis_vocabularies.InstitutionEventRelation'),
        ),
        migrations.AddField(
            model_name='event',
            name='kind',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='apis_vocabularies.EventType'),
        ),
        migrations.AddField(
            model_name='event',
            name='person_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='event_set', through='apis_relations.PersonEvent', to='apis_vocabularies.PersonEventRelation'),
        ),
        migrations.AddField(
            model_name='event',
            name='place_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='event_set', through='apis_relations.PlaceEvent', to='apis_vocabularies.PlaceEventRelation'),
        ),
        migrations.AddField(
            model_name='event',
            name='work_relationtype_set',
            field=models.ManyToManyField(blank=True, related_name='event_set', through='apis_relations.EventWork', to='apis_vocabularies.EventWorkRelation'),
        ),
        migrations.AddField(
            model_name='event',
            name='work_set',
            field=models.ManyToManyField(blank=True, related_name='event_set', through='apis_relations.EventWork', to='apis_entities.Work'),
        ),
    ]



from crispy_forms.helper import FormHelper
from django import forms

from .models import Label


# TODO __sresch__ : inspect if this should be removed or adapted
# since this class seems redundant as the same functionality is provided in apis_relations/forms.py in class EntityLabelForm there
#
# This class LabelForm here is only refered to in apis_labels/views.py which would be loaded by the respective urls in apis_labels/urls.py
# However none of these urls seem to work when called in the browser
class LabelForm(forms.ModelForm):
    isoCode = forms.CharField(
        required=True,
        label="ISO 639-3"
    )
    label = forms.CharField(required=True, help_text="The entities label or name.")

    class Meta:
        model = Label
        fields = ['label', 'isoCode_639_3', 'label_type']

    def __init__(self, *args, **kwargs):
        super(LabelForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = False



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse
from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient
from django.contrib.contenttypes.models import ContentType
from rest_framework import status

from .models import Place, Person


class PersonModelTestCase(TestCase):
    name = "test name"
    first_name = "test first name"
    place_name = "test place name"
    start_date = "1.3.1930"
    end_date = "4.6.1960"

    @classmethod
    def setUpTestData(cls):
        # Set up data for the whole TestCase
        cls.person = Person.objects.create(name=cls.name, first_name=cls.first_name)
        cls.place = Place.objects.create(name=cls.place_name)
        user = User.objects.create_user(username="lauren", password="pas_1234$")
        token = Token.objects.create(user=user)
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION="Token " + token.key)
        cls.c = client

    def test_apiget(self):
        token = Token.objects.get(user__username="lauren")
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION="Token " + token.key)
        res = client.get(
            reverse("apis:apis_core:person-detail", kwargs={"pk": self.person.pk})
        )
        print(res)


class GetTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set up data for the whole TestCase
        user = User.objects.create_user(username="lauren", password="pas_1234$")
        token = Token.objects.create(user=user)
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION="Token " + token.key)
        cls.c = client

    def test_get(self):
        # client = APIClient()
        for ent in ContentType.objects.filter(
            app_label__in=[
                "apis_entities",
                "apis_relations",
                "apis_metainfo",
                "apis_vocabularies",
            ]
        ).exclude(model__in=["tempentityclass", "relationbaseclass"]):
            print(f"Testing GET for: {ent.model.lower().replace(' ', '')}")
            url = (
                reverse(f'apis:apis_core:{ent.model.lower().replace(" ", "")}-list')
                + "?format=json"
            )
            res = self.c.get(url)
            self.assertEqual(res.status_code, status.HTTP_200_OK)
            d = res.json()
            if len(d["results"]) > 0:
                print(
                    f"Result of {ent.model.lower().replace(' ', '')} > 0, testing get of first element"
                )
                res2 = self.c.get(d["results"][0]["url"])
                self.assertEqual(res2.status_code, status.HTTP_200_OK)



from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.http import HttpResponse
from django.shortcuts import redirect, get_object_or_404
from django.template.loader import select_template
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.generic import DeleteView
from django_tables2 import RequestConfig
from guardian.core import ObjectPermissionChecker
from reversion.models import Version
import importlib

from apis_core.apis_entities.models import AbstractEntity
from apis_core.apis_labels.models import Label
from apis_core.apis_metainfo.models import Uri
from apis_core.apis_relations.models import AbstractRelation
from apis_core.apis_relations.tables import get_generic_relations_table, LabelTableEdit
from .forms import get_entities_form, FullTextForm, GenericEntitiesStanbolForm
from .views import get_highlighted_texts
from .views import set_session_variables
from ..apis_vocabularies.models import TextType

if 'apis_highlighter' in settings.INSTALLED_APPS:
    from apis_highlighter.forms import SelectAnnotatorAgreement


@method_decorator(login_required, name='dispatch')
class GenericEntitiesEditView(View):

    def get(self, request, *args, **kwargs):
        entity = kwargs['entity']
        pk = kwargs['pk']
        entity_model = AbstractEntity.get_entity_class_of_name(entity)
        instance = get_object_or_404(entity_model, pk=pk)
        request = set_session_variables(request)
        relations = AbstractRelation.get_relation_classes_of_entity_name(entity_name=entity)
        side_bar = []
        for rel in relations:
            match = [
                rel.get_related_entity_classA().__name__.lower(),
                rel.get_related_entity_classB().__name__.lower()
            ]
            prefix = "{}{}-".format(match[0].title()[:2], match[1].title()[:2])
            table = get_generic_relations_table(relation_class=rel, entity_instance=instance, detail=False)
            title_card = ''
            if match[0] == match[1]:
                title_card = entity.title()
                dict_1 = {'related_' + entity.lower() + 'A': instance}
                dict_2 = {'related_' + entity.lower() + 'B': instance}
                if 'apis_highlighter' in settings.INSTALLED_APPS:
                    objects = rel.objects.filter_ann_proj(request=request).filter(
                        Q(**dict_1) | Q(**dict_2))
                else:
                    objects = rel.objects.filter(
                        Q(**dict_1) | Q(**dict_2))
            else:
                if match[0].lower() == entity.lower():
                    title_card = match[1].title()
                else:
                    title_card = match[0].title()
                dict_1 = {'related_' + entity.lower(): instance}
                if 'apis_highlighter' in settings.INSTALLED_APPS:
                    objects = rel.objects.filter_ann_proj(request=request).filter(**dict_1)
                else:
                    objects = rel.objects.filter(**dict_1)
            tb_object = table(data=objects, prefix=prefix)
            tb_object_open = request.GET.get(prefix + 'page', None)
            RequestConfig(request, paginate={"per_page": 10}).configure(tb_object)
            side_bar.append((title_card, tb_object, ''.join([x.title() for x in match]), tb_object_open))
        form = get_entities_form(entity.title())
        form = form(instance=instance)
        form_text = FullTextForm(entity=entity.title(), instance=instance)
        if 'apis_highlighter' in settings.INSTALLED_APPS:
            form_ann_agreement = SelectAnnotatorAgreement()
        else:
            form_ann_agreement = False
        if 'apis_bibsonomy' in settings.INSTALLED_APPS:
            apis_bibsonomy = getattr(settings, 'APIS_BIBSONOMY_FIELDS', ['self'])
            apis_bibsonomy_texts = getattr(settings, "APIS_BIBSONOMY_TEXTS", False)
            if apis_bibsonomy_texts:
                apis_bibsonomy.extend([f"text_{pk}" for pk in TextType.objects.filter(name__in=apis_bibsonomy_texts).values_list('pk', flat=True) if f"text_{pk}" not in apis_bibsonomy])
            if isinstance(apis_bibsonomy, list):
                apis_bibsonomy = '|'.join([x.strip() for x in apis_bibsonomy])
        else:
            apis_bibsonomy = False
        object_revisions = Version.objects.get_for_object(instance)
        object_lod = Uri.objects.filter(entity=instance)
        object_texts, ann_proj_form = get_highlighted_texts(request, instance)
        object_labels = Label.objects.filter(temp_entity=instance)
        tb_label = LabelTableEdit(data=object_labels, prefix=entity.title()[:2] + 'L-')
        tb_label_open = request.GET.get('PL-page', None)
        side_bar.append(('Label', tb_label, 'PersonLabel', tb_label_open))
        RequestConfig(request, paginate={"per_page": 10}).configure(tb_label)
        perm = ObjectPermissionChecker(request.user)
        permissions = {'change': perm.has_perm('change_{}'.format(entity), instance),
                       'delete': perm.has_perm('delete_{}'.format(entity), instance),
                       'create': request.user.has_perm('entities.add_{}'.format(entity))}
        template = select_template(['apis_entities/{}_create_generic.html'.format(entity),
                                    'apis_entities/entity_create_generic.html'])
        context = {
            'entity_type': entity,
            'form': form,
            'form_text': form_text,
            'instance': instance,
            'right_card': side_bar,
            'object_revisions': object_revisions,
            'object_texts': object_texts,
            'object_lod': object_lod,
            'ann_proj_form': ann_proj_form,
            'form_ann_agreement': form_ann_agreement,
            'apis_bibsonomy': apis_bibsonomy,
            'permissions': permissions}
        form_merge_with = GenericEntitiesStanbolForm(entity, ent_merge_pk=pk)
        context['form_merge_with'] = form_merge_with
        return HttpResponse(template.render(request=request, context=context))

    def post(self, request, *args, **kwargs):
        entity = kwargs['entity']
        pk = kwargs['pk']
        entity_model = AbstractEntity.get_entity_class_of_name(entity)
        instance = get_object_or_404(entity_model, pk=pk)
        form = get_entities_form(entity.title())
        form = form(request.POST, instance=instance)
        form_text = FullTextForm(request.POST, entity=entity.title())
        if form.is_valid() and form_text.is_valid():
            entity_2 = form.save()
            form_text.save(entity_2)
            return redirect(reverse('apis:apis_entities:generic_entities_edit_view', kwargs={
                'pk': pk, 'entity': entity
            }))
        else:
            template = select_template(['apis_entities/{}_create_generic.html'.format(entity),
                                        'apis_entities/entity_create_generic.html'])
            perm = ObjectPermissionChecker(request.user)
            permissions = {'change': perm.has_perm('change_{}'.format(entity), instance),
                           'delete': perm.has_perm('delete_{}'.format(entity), instance),
                           'create': request.user.has_perm('entities.add_{}'.format(entity))}
            context = {
                'form': form,
                'entity_type': entity,
                'form_text': form_text,
                'instance': instance,
                'permissions': permissions}
            if entity.lower() != 'place':
                form_merge_with = GenericEntitiesStanbolForm(entity, ent_merge_pk=pk)
                context['form_merge_with'] = form_merge_with
                return TemplateResponse(request, template, context=context)
            return HttpResponse(template.render(request=request, context=context))


@method_decorator(login_required, name='dispatch')
class GenericEntitiesCreateView(View):
    def get(self, request, *args, **kwargs):
        entity = kwargs['entity']
        form = get_entities_form(entity.title())
        form = form()
        form_text = FullTextForm(entity=entity.title())
        permissions = {'create': request.user.has_perm('entities.add_{}'.format(entity))}
        template = select_template(['apis_entities/{}_create_generic.html'.format(entity),
                                    'apis_entities/entity_create_generic.html'])
        return HttpResponse(template.render(request=request, context={
            'entity_type': entity,
            'permissions': permissions,
            'form': form,
            'form_text': form_text}))

    def post(self, request, *args, **kwargs):
        entity = kwargs['entity']
        form = get_entities_form(entity.title())
        form = form(request.POST)
        form_text = FullTextForm(request.POST, entity=entity.title())
        if form.is_valid() and form_text.is_valid():
            entity_2 = form.save()
            form_text.save(entity_2)
            return redirect(reverse('apis:apis_entities:generic_entities_detail_view', kwargs={
                'pk': entity_2.pk, 'entity': entity
            }))
        else:
            permissions = {'create': request.user.has_perm('apis_entities.add_{}'.format(entity))}
            template = select_template(['apis_entities/{}_create_generic.html'.format(entity),
                                        'apis_entities/entity_create_generic.html'])
            return HttpResponse(template.render(request=request, context={
                'permissions': permissions,
                'form': form,
                'form_text': form_text}))


@method_decorator(login_required, name='dispatch')
class GenericEntitiesCreateStanbolView(View):

    def post(self, request, *args, **kwargs):
        entity = kwargs['entity']
        ent_merge_pk = kwargs.get('ent_merge_pk', False)
        if ent_merge_pk:
            form = GenericEntitiesStanbolForm(entity, request.POST, ent_merge_pk=ent_merge_pk)
        else:
            form = GenericEntitiesStanbolForm(entity, request.POST)
        #form = form(request.POST)
        if form.is_valid():
            entity_2 = form.save()
            if ent_merge_pk:
                entity_2.merge_with(int(ent_merge_pk))
            return redirect(reverse('apis:apis_entities:generic_entities_edit_view', kwargs={
                'pk': entity_2.pk, 'entity': entity
            }))
        else:
            permissions = {'create': request.user.has_perm('apis_entities.add_{}'.format(entity))}
            template = select_template(['apis_entities/{}_create_generic.html'.format(entity),
                                        'apis_entities/entity_create_generic.html'])
            return HttpResponse(template.render(request=request, context={
                'permissions': permissions,
                'form': form}))


@method_decorator(login_required, name='dispatch')
class GenericEntitiesDeleteView(DeleteView):
    model = importlib.import_module("apis_core.apis_metainfo.models").TempEntityClass
    template_name = getattr(
        settings, 'APIS_DELETE_VIEW_TEMPLATE', 'apis_entities/confirm_delete.html'
    )

    def dispatch(self, request, *args, **kwargs):
        entity = kwargs['entity']
        self.success_url = reverse(
            'apis_core:apis_entities:generic_entities_list', kwargs={'entity': entity}
        )
        return super(GenericEntitiesDeleteView, self).dispatch(request, *args, **kwargs)
